"""
Main module for handling and instantiating synaptical connections and gap-junctions
"""
from __future__ import absolute_import
import hashlib
import logging
import numpy
from collections import defaultdict, Counter
from itertools import chain
from os import path as ospath
from typing import List, Optional

from .core import NeurodamusCore as Nd
from .core import ProgressBarRank0 as ProgressBar, MPI
from .core import run_only_rank0
from .core.configuration import GlobalConfig, SimConfig, ConfigurationError, find_input_file
from .connection import Connection, ReplayMode
from .io.synapse_reader import SynapseReader
from .target_manager import TargetManager, TargetSpec
from .utils import compat, bin_search, dict_filter_map
from .utils.logging import VERBOSE_LOGLEVEL, log_verbose, log_all
from .utils.memory import DryRunStats
from .utils.timeit import timeit
from .utils.pyutils import gen_ranges


class ConnectionSet(object):
    """
    A dataset of connections.
    Several populations may exist with different seeds
    """

    def __init__(self, src_id, dst_id, conn_factory=Connection):
        # Connections indexed by post-gid, then ordered by pre-gid
        self.src_id = src_id
        self.dst_id = dst_id
        self.src_name = None
        self.dst_name = None
        self.virtual_source = False
        self._conn_factory = conn_factory
        self._connections_map = defaultdict(list)
        self._conn_count = 0

    def __contains__(self, item):
        return item in self._connections_map

    def __getitem__(self, item):
        return self._connections_map[item]

    def get(self, item):
        return self._connections_map.get(item)

    def items(self):
        """Iterate over the population as tuples (dst_gid, [connections])"""
        return self._connections_map.items()

    def target_gids(self):
        """Get the list of all targets gids in this Population"""
        return self._connections_map.keys()

    def all_connections(self):
        """Get an iterator over all the connections."""
        return chain.from_iterable(self._connections_map.values())

    def _find_connection(self, sgid, tgid, exact=True):
        """Finds a connection, given its source and destination gids.

        Returns:
            tuple: connection list and index.
                If the element doesnt exist, index depends on 'exact':
                None if exact=True, otherwise the insertion index.
        """
        cell_conns = self._connections_map[tgid]
        pos = 0
        if cell_conns:
            pos = bin_search(cell_conns, sgid, lambda x: x.sgid)
        if exact and (pos == len(cell_conns) or cell_conns[pos].sgid != sgid):
            # Not found
            return cell_conns, None
        return cell_conns, pos

    # -
    def get_connection(self, sgid, tgid):
        """Retrieves a connection from the pre and post gids.

        Returns:
            Connection: A connection object if it exists. None otherwise
        """
        conn_lst, idx = self._find_connection(sgid, tgid)
        return None if idx is None else conn_lst[idx]

    # -
    def store_connection(self, conn):
        """When we have created a new connection (sgid->tgid), store it
        in order in our structure for faster retrieval later

        Args:
            conn: The connection object to be stored
        """
        cell_conns, pos = self._find_connection(conn.sgid, conn.tgid, exact=False)
        if cell_conns and pos < len(cell_conns) and cell_conns[pos].sgid == conn.sgid:
            logging.error("Attempt to store existing connection: %d->%d",
                          conn.sgid, conn.tgid)
            return
        self._conn_count += 1
        cell_conns.insert(pos, conn)

    # -
    def get_or_create_connection(self, sgid, tgid, **kwargs):
        """Returns a connection by pre-post gid, creating if required."""
        conns = self._connections_map[tgid]
        pos = 0
        if conns:
            # optimize for ordered insertion, and handle when sgid is not used
            last_conn = conns[-1]
            if last_conn.sgid in (sgid, None):
                return last_conn
            if last_conn.sgid < sgid:
                pos = len(conns)
            else:
                pos = bin_search(conns, sgid, lambda x: x.sgid)
                if conns[pos].sgid == sgid:
                    return conns[pos]
        # Not found. Create & insert
        cur_conn = self._conn_factory(sgid, tgid, self.src_id, self.dst_id, **kwargs)
        conns.insert(pos, cur_conn)
        self._conn_count += 1
        return cur_conn

    # -
    def get_connections(self, post_gids, pre_gids=None):
        """Get all connections between groups of gids."""
        if isinstance(post_gids, int):
            if pre_gids is None:
                return self._connections_map[post_gids]
            elif isinstance(pre_gids, int):
                elem = self.get_connection(pre_gids, post_gids)
                return (elem,) if elem is not None else ()

        post_gid_conn_lists = (
            self._connections_map.values() if post_gids is None
            else (self._connections_map[post_gids],) if isinstance(post_gids, int)
            else (self._connections_map[tgid] for tgid in post_gids)
        )
        if pre_gids is None:
            return chain.from_iterable(post_gid_conn_lists)
        elif isinstance(pre_gids, int):
            # Return a generator which is employing bin search
            return (conns[posi] for conns in post_gid_conn_lists
                    for posi in (bin_search(conns, pre_gids, lambda x: x.sgid),)
                    if posi < len(conns) and conns[posi].sgid == pre_gids)
        else:
            # Generic case. Looks through all conns in selected tgids
            pre_gids = set(pre_gids)
            return (c for conns in post_gid_conn_lists
                    for c in conns
                    if c.sgid in pre_gids)

    def get_synapse_params_gid(self, target_gid):
        """Get an iterator over all the synapse parameters of a target
        cell connections.
        """
        conns = self._connections_map[target_gid]
        return chain.from_iterable(c.synapse_params for c in conns)

    def delete(self, sgid, tgid):
        """Removes a given connection from the population."""
        conn_lst, idx = self._find_connection(sgid, tgid)
        if idx is None:
            logging.error("Non-existing connection to delete: %d->%d", sgid, tgid)
            return
        self._conn_count -= 1
        del conn_lst[idx]

    def delete_group(self, post_gids, pre_gids=None):
        """Removes a set of connections from the population."""
        for conns, indices in self._find_connections(post_gids, pre_gids):
            conns[:] = numpy.delete(conns, indices, axis=0).tolist()
            self._conn_count -= len(indices)

    def count(self):
        return self._conn_count

    # -
    def _find_connections(self, post_gids, pre_gids=None):
        """Get the indices of the connections between groups of gids"""
        post_gid_conn_lists = (
            self._connections_map.values() if post_gids is None
            else (self._connections_map[post_gids],) if isinstance(post_gids, int)
            else (self._connections_map[tgid] for tgid in post_gids)
        )

        if pre_gids is None:
            return ((conns, range(len(conns))) for conns in post_gid_conn_lists)

        sgids_interest = [pre_gids] if isinstance(pre_gids, int) else pre_gids
        return (
            (conns, numpy.searchsorted(
                numpy.fromiter((c.sgid for c in conns), dtype="int64", count=len(conns)),
                sgids_interest))
            for conns in post_gid_conn_lists
        )

    def ids_match(self, population_ids, dst_second=None):
        """Whereas a given population_id selector matches population
        """
        if isinstance(population_ids, tuple):
            expr_src, expr_dst = population_ids
        else:
            expr_src, expr_dst = (population_ids, dst_second)
        return ((expr_src is None or expr_src == self.src_id) and
                (expr_dst is None or expr_dst == self.dst_id))

    def is_default(self):
        return self.src_id == 0 and self.dst_id == 0

    def __str__(self):
        if self.is_default():
            return "<ConnectionSet: Default>"
        return "<ConnectionSet: %d-%d (%s->%s)>" % (
               self.src_id, self.dst_id, self.src_name, self.dst_name)

    def __repr__(self):
        return str(self)


class ConnectionManagerBase(object):
    """
    An abstract base class common to Synapse and GapJunction connections

    Connection Managers hold and manage connectivity among cell populations.
    For every src-dst pop pairs a new ConnectionManahger is created. The only case
    it holds several ConnectionSets is for old-style projections (no population names)
   """

    CONNECTIONS_TYPE = None
    """The type of connections subclasses handle"""

    # Set depending Classes, customizable
    ConnectionSet = ConnectionSet
    SynapseReader = SynapseReader
    conn_factory = Connection

    cell_manager = property(lambda self: self._cell_manager)
    src_cell_manager = property(lambda self: self._src_cell_manager)
    is_file_open = property(lambda self: bool(self._synapse_reader))
    src_pop_offset = property(lambda self: self._src_cell_manager.local_nodes.offset)
    target_pop_offset = property(lambda self: self._cell_manager.local_nodes.offset)

    # -
    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Base class c-tor for connections (Synapses & Gap-Junctions)

        Args:
            circuit_conf: A circuit config object where to get the synapse source from
                or None if the ConnecionManager is to be constructed empty
            target_manager: A target manager, where to query target cells
            cell_manager: Manager to query the local target cells
            src_cell_manager: Manager to query the local source cells [default: None -> internal]
            load_offsets: Whether to load synapse offsets

        """
        self._target_manager: TargetManager = target_manager
        self._cell_manager = cell_manager
        self._src_cell_manager = src_cell_manager or cell_manager

        self._populations = {}  # Multiple edge populations support. key is a tuple (src, dst)
        self._cur_population = None
        self._disabled_conns = defaultdict(list)

        self._synapse_reader = None
        self._raw_gids = cell_manager.local_nodes.raw_gids()
        self._total_connections = 0
        self.circuit_conf = circuit_conf
        self._src_target_filter = None  # filter by src target in all_connect (E.g: GapJ)
        # Load offsets might be required globally. Conn managers shall honor this if possible
        self._load_offsets = kw.get("load_offsets", False)
        # An internal var to enable collection of synapse statistics to a Counter
        self._dry_run_stats: DryRunStats = kw.get("dry_run_stats")
        self._dry_run_counted_cells = set()

    def __str__(self):
        return "<{:s} | {:s} -> {:s}>".format(
            self.__class__.__name__, str(self._src_cell_manager), str(self._cell_manager))

    def open_edge_location(self, syn_source, circuit_conf, **kw):
        edge_file, *pop = syn_source.split(":")
        pop_name = pop[0] if pop else None
        n_files = circuit_conf.get("NumSynapseFiles")
        src_pop_id = _get_projection_population_id(circuit_conf)
        return self.open_synapse_file(edge_file, pop_name, n_files, src_pop_id=src_pop_id, **kw)

    def open_synapse_file(self, synapse_file, edge_population, n_files=1, *,
                          src_pop_id=None, src_name=None, **_kw):
        """Initializes a reader for Synapses config objects and associated population

        Args:
            synapse_file: The nrn/edge file. For old nrn files it may be a dir.
            edge_population: The population of the edges
            n_files: (nrn only) the number of nrn files in the directory (without nrn.h5)
            src_pop_id: (compat) Allow overriding the src population ID
            src_name: The source pop name, normally matching that of the source cell manager
        """
        if not ospath.isabs(synapse_file):
            synapse_file = find_input_file(synapse_file)
            log_verbose("Relative path for edge file resolved to: %s", synapse_file)
        if not ospath.exists(synapse_file):
            raise ConfigurationError("Connectivity (Edge) file not found: {}".format(synapse_file))
        if ospath.isdir(synapse_file):
            logging.warning("Edges source is a directory (legacy nrn.h5 only)")
            if ospath.isfile(ospath.join(synapse_file, "nrn.h5")):
                n_files = 1

        self._synapse_reader = self._open_synapse_file(synapse_file, edge_population, n_files)
        if self._load_offsets:
            if not self._synapse_reader.has_property("synapse_index"):
                raise Exception("Synapse offsets required but not available. "
                                "Please use a more recent version of neurodamus-core/synapse-tool")

        self._init_conn_population(src_name, src_pop_id)
        self._unlock_all_connections()  # Allow appending synapses from new sources
        return synapse_file

    # - override if needed
    def _open_synapse_file(self, synapse_file, pop_name, n_nrn_files=None):
        logging.debug("Opening Synapse file %s, population: %s", synapse_file, pop_name)
        return self.SynapseReader.create(
            synapse_file, self.CONNECTIONS_TYPE, pop_name,
            n_nrn_files, self._raw_gids,  # Used eventually by NRN reader
            extracellular_calcium=SimConfig.extracellular_calcium
        )

    def _init_conn_population(self, src_pop_name, pop_id_override):
        if not src_pop_name:
            src_pop_name = self._src_cell_manager.population_name
        dst_pop_name = self._cell_manager.population_name
        src_pop_id, dst_pop_id = self._compute_pop_ids(src_pop_name, dst_pop_name, pop_id_override)

        if self._cur_population and src_pop_id == 0 and not src_pop_name:
            logging.warning("Neither Sonata population nor populationID set. "
                            "Edges will be merged with base circuit")

        cur_pop = self.select_connection_set(src_pop_id, dst_pop_id)  # type: ConnectionSet
        cur_pop.src_name = src_pop_name
        cur_pop.dst_name = dst_pop_name
        cur_pop.virtual_source = (self._src_cell_manager.is_virtual
                                  or src_pop_name != self._src_cell_manager.population_name
                                  or bool(pop_id_override) and not src_pop_name)
        logging.info("Loading connections to population: %s", cur_pop)

    def _compute_pop_ids(self, src_pop, dst_pop, src_pop_id=None):
        """Compute pop id automatically. pop src 0 is base population.
        if src_pop_id is provided, it will be used instead.
        """
        def make_id(node_pop):
            pop_hash = hashlib.md5(node_pop.encode()).digest()
            return ((pop_hash[1] & 0x0f) << 8) + pop_hash[0]  # id: 12bit hash

        dst_pop_id = 0 if self._cell_manager.is_default else make_id(dst_pop)
        if src_pop_id is None:
            src_pop_id = 0 if self._src_cell_manager.is_default else make_id(src_pop)
        return src_pop_id, dst_pop_id

    # -
    def select_connection_set(self, src_pop_id, dst_pop_id):
        """Select the active population of connections given src and dst node pop ids.
        `connect_all()` and `connect_group()` will apply only to the active population.

        Returns: The selected ConnectionSet, eventually created
        """
        self._cur_population = self.get_population(src_pop_id, dst_pop_id)
        return self._cur_population

    # -
    def get_population(self, src_pop_id, dst_pop_id=0):
        """Retrieves a connection set given node src and dst pop ids"""
        pop = self._populations.get((src_pop_id, dst_pop_id))
        if not pop:
            pop = self.ConnectionSet(src_pop_id, dst_pop_id, conn_factory=self.conn_factory)
            self._populations[(src_pop_id, dst_pop_id)] = pop
        return pop

    # NOTE: Several methods use a selector of the connectivity populations
    # which, to be backwards compat, can be a single ID of the src_population
    # or a tuple to specify source and destination

    def find_populations(self, population_ids):
        """Finds the populations that match a given population selector.

        Args:
            population_ids: A population ids selector. Accepted formats:

                - None: All
                - int: selects matching source population id.
                - tuple(src: Any, dst: Any): Specify source and dest.
                  Each value can also be None, e.g.: (None, 1) selects all
                  populations having post id 1
        """
        if (isinstance(population_ids, tuple)
                and population_ids[0] is not None
                and population_ids[1] is not None):
            return [self._populations[population_ids]]
        return [
            pop for pop in self._populations.values()
            if pop.ids_match(population_ids)
        ]

    # -
    def all_connections(self):
        """Retrieves all the existing connections"""
        return chain.from_iterable(
            pop.all_connections() for pop in self._populations.values())

    @property
    def connection_count(self):
        return self._total_connections

    @property
    def current_population(self):
        return self._cur_population

    # -
    def get_connections(self, post_gids, pre_gids=None, population_ids=None):
        """Retrieves all connections that match post and pre gids eventually
        in a subset of the populations.

        Note: Due to multi-population, a connection may not be unique
        for a given pre-post gid pair. As such get_connection() doesn't
        make sense anymore and this method shall be used instead.

        Args:
            post_gids: The target gids of the connections to search (None = All)
            pre_gids: idem for pre-gids. [Default: all)
            population_ids: A int/tuple of populations ids. Default: all

        """
        for pop in self.find_populations(population_ids):
            for conn in pop.get_connections(post_gids, pre_gids):
                yield conn

    # -
    def create_connections(self, src_target=None, dst_target=None):
        """Creates connections according to loaded parameters in 'Connection'
        blocks of the BlueConfig in the currently active ConnectionSet.

        If no Connection block relates to the current population, then load all
        edges. If a single blocks exists with Weight=0, skip creation entirely.

        NOTE: All connections respecting the src_target are retrieved
        and created, even if they use src cells which are NOT instantiated.
        This is to support replay and other stimulus which dont need the src cell.
        If only a subset of connections is wanted, they can be filtered by specifying
        the "source" attribute of the respective connection blocks.

        Args:
            src_target: Target name to restrict creating connections coming from it
            dst_target: Target name to restrict creating connections going into it
        """
        conn_src_spec = TargetSpec(src_target)  # instantiate all from src
        conn_src_spec.population = self.current_population.src_name
        conn_dst_spec = TargetSpec(dst_target or self.cell_manager.circuit_target)
        conn_dst_spec.population = self.current_population.dst_name
        this_pathway = {
            "Source": str(conn_src_spec),
            "Destination": str(conn_dst_spec)
        }
        matching_conns = [
            conn for conn in SimConfig.connections.values()
            if self._target_manager.pathways_overlap(conn, this_pathway)
        ]
        if not matching_conns:
            logging.info("No matching Connection blocks. Loading all synapses...")
            self.connect_all()
            return

        # if we have a single connect block with weight=0, skip synapse creation entirely
        if len(matching_conns) == 1 and matching_conns[0].get("Weight") == .0:
            logging.warning("SKIPPING Connection create since they have invariably weight=0")
            return

        logging.info("Creating group connections (%d groups match)", len(matching_conns))
        for conn_conf in matching_conns:
            if "Delay" in conn_conf and conn_conf["Delay"] > 0:
                # Delayed connections are for configuration only, not creation
                continue

            # check if we are not supposed to create (only configure later)
            if conn_conf.get("CreateMode") == "NoCreate":
                continue

            conn_src = conn_conf["Source"]
            conn_dst = conn_conf["Destination"]
            synapse_id = conn_conf.get("SynapseID")
            mod_override = conn_conf.get("ModOverride")
            self.connect_group(conn_src, conn_dst, synapse_id, mod_override)

    # -
    def configure_connections(self, conn_conf):
        """Configure-only circuit connections according to a BlueConfig Connection block

        Args:
            conn_conf: The BlueConfig configuration block (dict)
        """
        log_msg = " * Pathway {:s} -> {:s}".format(conn_conf["Source"], conn_conf["Destination"])

        if "Delay" in conn_conf and conn_conf["Delay"] > 0:
            log_msg += ":\t[DELAYED] t={0[Delay]:g}, weight={0[Weight]:g}".format(conn_conf)
            configured_conns = self.setup_delayed_connection(conn_conf)
        else:
            if "SynapseConfigure" in conn_conf:
                log_msg += ":\tconfigure with '{:s}'".format(conn_conf["SynapseConfigure"])
            if "NeuromodStrength" in conn_conf:
                log_msg += "\toverwrite NeuromodStrength = {:g}" \
                    .format(conn_conf["NeuromodStrength"])
            if "NeuromodDtc" in conn_conf:
                log_msg += "\toverwrite NeuromodDtc = {:g}".format(conn_conf["NeuromodDtc"])
            configured_conns = self.configure_group(conn_conf)

        all_ranks_total = MPI.allreduce(configured_conns, MPI.SUM)
        if all_ranks_total > 0:
            logging.info(log_msg)
            logging.info(" => Configured {:g} connections".format(all_ranks_total))

    def setup_delayed_connection(self, conn_config):
        raise NotImplementedError("Manager %s doesn't implement delayed connections"
                                  % self.__class__.__name__)

    # -
    def connect_all(self, weight_factor=1, only_gids=None):
        """For every gid access its synapse parameters and instantiate
        all synapses.

        Args:
            weight_factor: Factor to scale all netcon weights (default: 1)
            only_gids: Create connections only for these tgids (default: Off)
        """
        if SimConfig.dry_run:
            counts = self._get_conn_stats(None)
            log_all(VERBOSE_LOGLEVEL, "[Rank %d] Synapse count: %d", MPI.rank, sum(counts.values()))
            self._dry_run_stats.synapse_counts += counts
            return

        conn_options = {'weight_factor': weight_factor}
        pop = self._cur_population

        for sgid, tgid, syns_params, extra_params, offset in \
                self._iterate_conn_params(self._src_target_filter, None, only_gids, True):
            if self._load_offsets:
                conn_options["synapses_offset"] = extra_params["synapse_index"][0]
            # Create all synapses. No need to lock since the whole file is consumed
            cur_conn = pop.get_or_create_connection(sgid, tgid, **conn_options)
            self._add_synapses(cur_conn, syns_params, None, offset)

    # -
    def connect_group(self, conn_source, conn_destination, synapse_type_restrict=None,
                      mod_override=None):
        """Instantiates pathway connections & synapses given src-dst

        Args:
            conn_source (str): The target name of the source cells
            conn_destination (str): The target of the destination cells
            synapse_type_restrict(int): Create only given synType synapses
            mod_override (str): ModOverride given for this connection group
        """
        conn_kwargs = {}
        pop = self._cur_population
        logging.debug("Connecting group %s -> %s", conn_source, conn_destination)
        src_tspec = TargetSpec(conn_source)
        dst_tspec = TargetSpec(conn_destination)
        src_target = src_tspec.name and self._target_manager.get_target(src_tspec)
        dst_target = dst_tspec.name and self._target_manager.get_target(dst_tspec)

        if src_target and src_target.is_void() or dst_target and dst_target.is_void():
            logging.debug("Skip void connectivity for current connectivity: %s - %s",
                          conn_source, conn_destination)
            return

        if SimConfig.dry_run:
            counts = self._get_conn_stats(dst_target)
            count_sum = sum(counts.values())
            log_all(VERBOSE_LOGLEVEL, "%s -> %s: %d", pop.src_name, conn_destination, count_sum)
            self._dry_run_stats.synapse_counts += counts
            return

        for sgid, tgid, syns_params, extra_params, offset in \
                self._iterate_conn_params(src_target, dst_target, mod_override=mod_override):
            if sgid == tgid:
                logging.warning("Making connection within same Gid: %d", sgid)
            if self._load_offsets:
                conn_kwargs["synapses_offset"] = extra_params["synapse_index"][0]

            cur_conn = pop.get_or_create_connection(sgid, tgid, **conn_kwargs)
            if cur_conn.locked:
                continue
            self._add_synapses(cur_conn, syns_params, synapse_type_restrict, offset)
            cur_conn.locked = True

    # -
    def _add_synapses(self, cur_conn, syns_params, syn_type_restrict=None, base_id=0):
        if syn_type_restrict:
            syns_params = syns_params[syns_params['synType'] != syn_type_restrict]

        cur_conn.add_synapses(self._target_manager, syns_params, base_id)

    # -
    def get_updated_population_offsets(self, src_target, dst_target):
        sgid_offset = self._src_cell_manager.local_nodes.offset
        tgid_offset = self._cell_manager.local_nodes.offset
        if src_target:
            src_target.set_offset(sgid_offset)
        if dst_target:
            dst_target.set_offset(tgid_offset)
        return sgid_offset, tgid_offset

    # -
    class ConnDebugger:
        __slots__ = ['yielded_src_gids']

        def __init__(self):
            self.yielded_src_gids = compat.array("i") if GlobalConfig.debug_conn else None

        def register(self, sgid, base_tgid, syns_params):
            if not (debug_conn := GlobalConfig.debug_conn):
                return
            if debug_conn == [base_tgid]:
                self.yielded_src_gids.append(sgid)
            elif debug_conn == [sgid, base_tgid]:
                log_all(logging.DEBUG, "Connection (%d-%d). Params:\n%s",
                        sgid, base_tgid, syns_params)

        def __del__(self):
            if self.yielded_src_gids:
                log_all(logging.DEBUG, "Source GIDs for debug cell: %s", self.yielded_src_gids)

    # -
    def _iterate_conn_params(self, src_target, dst_target, gids=None, show_progress=False,
                             mod_override=None):
        """A generator which loads synapse data and yields tuples(sgid, tgid, synapses)

        Args:
            src_target: the target to filter the source cells, or None
            dst_target: the target to filter the destination cells, or None
            gids: Use given gids, instead of the circuit target cells. Default: None
            show_progress: Display a progress bar as tgids are processed
        """
        if src_target and src_target.is_void() or dst_target and dst_target.is_void():
            return

        def target_gids(gids):
            if gids is None:
                return self._raw_gids
            gids = numpy.intersect1d(gids, self._raw_gids)
            if dst_target:
                gids = numpy.intersect1d(gids, dst_target.get_raw_gids())
            return gids

        gids = target_gids(gids)
        created_conns_0 = self._cur_population.count()
        sgid_offset, tgid_offset = self.get_updated_population_offsets(src_target, dst_target)

        self._synapse_reader.configure_override(mod_override)
        self._synapse_reader.preload_data(gids)
        extra_fields = {}  # Without extra fields, reuse this object

        # NOTE: This routine is quite critical, sitting at the core of synapse processing
        # so it has been carefully optimized with numpy vectorized operations, even if
        # it might lose some readability.
        # For each tgid we obtain the synapse parameters as a record array. We then split it,
        # without copying, yielding ranges (views) of it.

        gids = ProgressBar.iter(gids) if show_progress else gids

        for base_tgid in gids:
            tgid = base_tgid + tgid_offset
            syns_params = self._synapse_reader.get_synapse_parameters(base_tgid)
            logging.debug("GID %d Syn count: %d", tgid, len(syns_params))

            if self._load_offsets:
                syn_index = self._synapse_reader.get_property(base_tgid, "synapse_index")
                extra_fields = {"synapse_index": syn_index}

            # We yield ranges of contiguous parameters belonging to the same connection,
            # and given we have data for a single tgid, enough to group by sgid.
            # The first row of a range is found by numpy.diff

            sgids = syns_params[syns_params.dtype.names[0]].astype("int64")  # src gid in field 0
            sgids_ranges = numpy.diff(sgids, prepend=numpy.nan, append=numpy.nan).nonzero()[0]
            conn_count = len(sgids_ranges) - 1
            conn_debugger = self.ConnDebugger()

            if src_target:
                # create a set with the gids that belong both to the synapses and the target
                unique_sgids = sgids[sgids_ranges[:-1]]
                allowed_sgids = set(unique_sgids[src_target.contains(unique_sgids, raw_gids=True)])
                n_yielded_conns = len(allowed_sgids)
                allowed_ranges = [(sgids_ranges[i], sgids_ranges[i+1]) for i in range(conn_count)
                                  if sgids[sgids_ranges[i]] in allowed_sgids]
            else:
                n_yielded_conns = conn_count
                allowed_ranges = [(sgids_ranges[i], sgids_ranges[i+1]) for i in range(conn_count)]

            for range_start, range_end in allowed_ranges:
                sgid = int(sgids[range_start])
                final_sgid = sgid + sgid_offset
                syn_params = syns_params[range_start:range_end]
                extra_params = extra_fields and {  # reuse empty {}. Dont modify later!
                    name: prop[range_start:range_end]
                    for name, prop in extra_fields.items()
                }
                conn_debugger.register(sgid, base_tgid, syn_params)
                yield final_sgid, tgid, syn_params, extra_params, range_start

            logging.debug(" > Yielded %d out of %d connections. (Filter by src Target: %s)",
                          n_yielded_conns, conn_count, src_target and src_target.name)

        created_conns = self._cur_population.count() - created_conns_0
        self._total_connections += created_conns

        all_created = MPI.allreduce(created_conns, MPI.SUM)
        if all_created:
            pathway_repr = "[ALL]"
            if src_target and dst_target:
                pathway_repr = "Pathway {} -> {}".format(src_target.name, dst_target.name)
            logging.info(" * %s. Created %d connections", pathway_repr, all_created)

    def _get_conn_stats(self, dst_target):
        """Estimates the number of synapses per type for the given destination target
        Note: measurement is done without any restriction on the source target.
        As so it can avoid counting the same synapse twice by keeping track of
        target cells alone.

        Args:
            dst_target: The target to estimate synapses for

        Returns:
            A Counter object with the estimated synapses per type

        Note:
          - _src target is not considered so we count all inbound synapses
          -  We will only consider gids which have not been accounted for yet.
        """
        BLOCK_BASE_SIZE = 5000
        SAMPLES_PER_BLOCK = 100

        # Get the raw gids for the destination target
        raw_gids = dst_target.get_local_gids(raw_gids=True) if dst_target else self._raw_gids

        # Filter out the gids which have been already considered
        # Use sets since they are much faster than numpy
        new_gids = set(raw_gids) - self._dry_run_counted_cells
        if not new_gids:
            # Either target is empty or all the cells were considered before
            return {}

        local_counter = Counter()

        # NOTE:
        #  - Estimation (and extrapolation) is performed per metype since properties can vary
        #  - Consider only the cells for the current target

        for metype, me_gids in self._dry_run_stats.metype_gids.items():
            me_gids = set(me_gids).intersection(new_gids)
            me_gids_count = len(me_gids)
            if not me_gids_count:
                continue

            logging.debug("Estimating synapses for metype %s", metype)
            self._dry_run_counted_cells.update(me_gids)  # track as seen
            me_gids = numpy.fromiter(me_gids, dtype="uint32")

            # NOTE:
            # Process the first 100 cells from increasingly large blocks
            #  - Takes advantage of data locality
            #  - Blocks increase as a geometric progression for handling very large sets

            metype_estimate = Counter()
            sampled_gids_count = 0

            for start, stop, in gen_ranges(me_gids_count, BLOCK_BASE_SIZE, block_increase_rate=1.1):
                logging.debug("Processing range %d:%d", start, stop)
                block_len = stop - start
                sample = me_gids[start:(start + SAMPLES_PER_BLOCK)]
                sample_len = len(sample)
                if not sample_len:
                    continue
                sample_counts = self._synapse_reader.get_counts(sample, group_by="syn_type_id")
                logging.debug("Gids: %s... Types: %s", sample[:10], sample_counts)
                logging.debug("Average syn/cell: %.2f", sum(sample_counts.values()) / sample_len)
                sampled_gids_count += sample_len
                ratio = block_len / sample_len
                metype_estimate.update({
                    syn_t: int(n * ratio)
                    for syn_t, n in sample_counts.items()}
                )

            # Extrapolation
            logging.debug("Cells samples / total: %d / %s", sampled_gids_count, me_gids_count)
            me_estimated_sum = sum(metype_estimate.values())
            log_all(VERBOSE_LOGLEVEL, "%s: Average syns/cell: %.1f, Estimated total: %d ",
                    metype, me_estimated_sum / me_gids_count, me_estimated_sum)
            local_counter.update(metype_estimate)

        return local_counter

    # -
    def get_target_connections(self, src_target_name,
                                     dst_target_name,
                                     selected_gids=None,
                                     conn_population=None):
        """Retrives the connections between src-dst cell targets

        Args:
             selected_gids: (optional) post gids to select (original, w/o offsetting)
             conn_population: restrict the set of connections to be returned
        """
        src_target_spec = TargetSpec(src_target_name)
        dst_target_spec = TargetSpec(dst_target_name)

        src_target = self._target_manager.get_target(src_target_spec) \
            if src_target_spec.name is not None else None
        assert dst_target_spec.name, "No target specified for `get_target_connections`"
        dst_target = self._target_manager.get_target(dst_target_spec)
        if src_target and src_target.is_void() or dst_target.is_void():
            return

        _, tgid_offset = self.get_updated_population_offsets(src_target, dst_target)
        populations: List[ConnectionSet] = (conn_population,) if conn_population is not None \
            else self._populations.values()

        # temporary set for faster lookup
        src_gids = src_target and set(src_target.get_gids())
        for population in populations:
            logging.debug("Connections from population %s", population)
            tgids = numpy.fromiter(population.target_gids(), 'uint32')
            tgids = numpy.intersect1d(tgids, dst_target.get_gids())
            if selected_gids:
                tgids = numpy.intersect1d(tgids, selected_gids + tgid_offset)
            for conn in population.get_connections(tgids):
                if src_target is None or conn.sgid in src_gids:
                    yield conn

    # -
    def configure_group(self, conn_config, gidvec=None):
        """Configure connections according to a BlueConfig Connection block

        Args:
            conn_config: The connection configuration dict
            gidvec: A restricted set of gids to configure (original, w/o offsetting)
        """
        src_target = conn_config["Source"]
        dst_target = conn_config["Destination"]
        _properties = {
            "Weight": "weight_factor",
            "SpontMinis": "minis_spont_rate",
            "SynDelayOverride": "syndelay_override",
            "NeuromodStrength": "neuromod_strength",
            "NeuromodDtc": "neuromod_dtc"
        }
        syn_params = dict_filter_map(conn_config, _properties)

        # Load eventual mod override helper
        if "ModOverride" in conn_config:
            logging.info("   => Overriding mod: %s", conn_config["ModOverride"])
            override_helper = conn_config["ModOverride"] + "Helper"
            Nd.load_hoc(override_helper)
            assert hasattr(Nd.h, override_helper), \
                "ModOverride helper doesn't define hoc template: " + override_helper

        configured_conns = 0
        for conn in self.get_target_connections(src_target, dst_target, gidvec):
            for key, val in syn_params.items():
                setattr(conn, key, val)
            if "ModOverride" in conn_config:
                conn.override_mod(conn_config.get('hoc') or compat.PyMap(conn_config).hoc_map)
            if "SynapseConfigure" in conn_config:
                conn.add_synapse_configuration(conn_config["SynapseConfigure"])
            configured_conns += 1
        return configured_conns

    # -
    def configure_group_delayed(self, conn_config, gidvec=None):
        """Update instantiated connections with configuration from a
        'Delayed Connection' blocks.
        """
        self.update_connections(conn_config["Source"],
                                conn_config["Destination"],
                                gidvec,
                                conn_config.get("SynapseConfigure"),
                                conn_config.get("Weight"))

    # Live connections update
    # -----------------------
    @timeit(name="connUpdate", verbose=False)
    def update_connections(self, src_target, dst_target, gidvec=None,
                                 syn_configure=None, weight=None, **syn_params):
        """Update params on connections that are already instantiated.

        Args:
            src_target: Name of Source Target
            dst_target: Name of Destination Target
            gidvec: A list of gids to apply configuration. Default: all
            syn_configure: A hoc configuration string to apply to the synapses
            weight: new weights for the netcons
            **syn_params: Keyword arguments of synapse properties to be changed
                e.g. conductance: g=xyz
        """
        if syn_configure is None and weight is None and not syn_params:
            logging.warning("No synpases parameters being updated for Targets %s->%s",
                            src_target, dst_target)
            return

        updated_conns = 0
        for conn in self.get_target_connections(src_target, dst_target, gidvec):
            if weight is not None:
                updated_conns += 1
                conn.update_weights(weight)
            if syn_configure is not None:
                conn.configure_synapses(syn_configure)
            if syn_params:
                conn.update_synpase_parameters(**syn_params)

        logging.info("Updated %d conns", updated_conns)

    def restart_events(self):
        """After restore, restart the artificial events (replay and spont minis)
        """
        for conn in self.all_connections():
            conn.restart_events()

    # ------------------------------------------------------------------
    # Delete, Disable / Enable
    # ------------------------------------------------------------------
    def delete(self, sgid, tgid, population_ids=None):
        """Deletes a connection given source and target gids.

        NOTE: Contrary to disable(), deleting a connection cannot be undone,
        however it may help saving computational resources.

        Args:
            sgid: The pre-gid of the cell
            tgid: The post-gid of the cell
            population_ids: The population selector. Default: all
        """
        for pop in self.find_populations(population_ids):
            pop.delete(sgid, tgid)

    def disable(self, sgid, tgid, also_zero_conductance=False, population_ids=None):
        """Disable a connection, its netcons and optionally synapses.

        NOTE: Disabling a connection before calling init() prevents
        it from being instantiated in CoreNeuron.

        Args:
            sgid: The pre-gid of the cell
            tgid: The post-gid of the cell
            also_zero_conductance: Also sets synapses' conductances
                to zero. Default: False
            population_ids: The population selector. Default: all
        """
        for pop in self.find_populations(population_ids):
            c = pop.get_connection(sgid, tgid)
            c.disable(also_zero_conductance)
            self._disabled_conns[tgid].append(c)

    def reenable(self, sgid, tgid, population_ids=None):
        """(Re)enable a connection from given populations.
        """
        allowed_pops = self.find_populations(population_ids)
        delete_indexes = []
        for i, conn in enumerate(self._disabled_conns[tgid]):
            if conn.sgid == sgid and any((p.src_id, p.dst_id) == conn.population_id
                                         for p in allowed_pops):
                conn.enable()
                delete_indexes.append(i)
        self._disabled_conns[tgid] = \
            numpy.delete(self._disabled_conns[tgid], delete_indexes).tolist()

    def reenable_all(self, post_gids=None):
        """Re-enables all disabled connections

        Args:
            post_gids: The list of target gids to enable (Default: all)
        """
        gids = self._raw_gids if post_gids is None else post_gids
        offset = self._cell_manager.local_nodes.offset
        for tgid in gids:
            tgid += offset
            for c in self._disabled_conns[tgid]:
                c.enable()
            del self._disabled_conns[tgid][:]

    # GROUPS
    # ------
    def delete_group(self, post_gids, pre_gids=None, population_ids=None):
        """Delete a number of connections given pre and post gid lists.
           Note: None is neutral and will match all gids.

        Args:
            post_gids: The target gids of the connections to be disabled
                Use None for all post-gids
            pre_gids: idem for pre-gids. [Default: None -> all)
            population_ids: A int/tuple of populations ids. Default: all
        """
        for pop in self.find_populations(population_ids):
            pop.delete_group(post_gids, pre_gids)

    def disable_group(self, post_gids, pre_gids=None,
                            also_zero_conductance=False,
                            population_ids=None):
        """Disable a number of connections given pre and post gid lists.

        Args:
            post_gids: The target gids of the connections to be deleted
            pre_gids: idem for pre-gids. [Default: None -> all)
            also_zero_conductance: Also sets synapse conductance to 0
            population_ids: A int/tuple of populations ids. Default: all
        """
        for pop in self.find_populations(population_ids):
            for conn in pop.get_connections(post_gids, pre_gids):
                self._disabled_conns[conn.tgid].append(conn)
                conn.disable(also_zero_conductance)

    def reenable_group(self, post_gids, pre_gids=None, population_ids=None):
        """Enable a number of connections given lists of pre and post gids.
        Note: None will match all gids.
        """
        if post_gids is None:
            post_gids = self._raw_gids
        offset = self._cell_manager.local_nodes.offset
        pre_gids = set(pre_gids)
        allowed_pops = self.find_populations(population_ids)

        for tgid in post_gids:
            tgid += offset
            to_delete = []
            for i, conn in enumerate(self._disabled_conns[tgid]):
                if conn.sgid in pre_gids and \
                        any((p.src_id, p.dst_id) == conn.population_id
                            for p in allowed_pops):
                    conn.enable()
                    to_delete.append(i)

            self._disabled_conns[tgid] = \
                numpy.delete(self._disabled_conns[tgid], to_delete).tolist()

    def get_disabled(self, post_gid=None):
        """Returns the list of disabled connections, optionally for a
        given post-gid.
        """
        if post_gid is not None:
            return self._disabled_conns[post_gid]
        return chain.from_iterable(self._disabled_conns.values())

    def _unlock_all_connections(self):
        """Unlock all, mainly when we load a new connectivity source"""
        for conn in self.all_connections():
            conn.locked = False

    def finalize(self, base_seed=0, sim_corenrn=False, *, _conn_type="synapses", **conn_params):
        """Instantiates the netcons and Synapses for all connections.

        Note: All weight scalars should have their final values.

        Args:
            base_seed: optional argument to adjust synapse RNGs (default=0)
            sim_corenrn: Finalize accordingly in case we target CoreNeuron
            _conn_type: (Internal) A string repr of the connectivity type
            conn_params: Additional finalize parameters for the specific _finalize_conns
                E.g. replay_mode (Default: Auto-Detect) Use DISABLED to skip replay
                and COMPLETE to instantiate VecStims in all synapses

        """
        self._synapse_reader = None  # Destroy to release memory (all cached params)
        logging.info("Instantiating %s... Params: %s", _conn_type, str(conn_params))
        n_created_conns = 0

        for popid, pop in self._populations.items():
            attach_src = (pop.src_id == 0 or not pop.virtual_source  # real populations
                          or pop.virtual_source and bool(sim_corenrn))  # Req'd for replay
            conn_params["attach_src_cell"] = attach_src
            logging.info(" * Connections among %s -> %s, attach src: %s",
                         pop.src_name or "(base)", pop.dst_name or "(base)", attach_src)

            for tgid, conns in ProgressBar.iter(pop.items(), name="Pop:" + str(popid)):
                n_created_conns += self._finalize_conns(
                    tgid, conns, base_seed, sim_corenrn, **conn_params)

        all_ranks_total = MPI.allreduce(n_created_conns, MPI.SUM)
        logging.info(" => Created %d %s", all_ranks_total, _conn_type)
        return all_ranks_total

    def _finalize_conns(self, tgid, conns, base_seed, sim_corenrn, *, reverse=False, **kwargs):
        """ Low-level handling of finalizing connections belonging to a target gid.
        By default it calls finalize on each cell.
        """
        # Note: *kwargs normally contains 'replay_mode' but may differ for other types
        metype = self._cell_manager.get_cell(tgid)
        n_created_conns = 0
        if reverse:
            conns = reversed(conns)
        for conn in conns:  # type: Connection
            syn_count = conn.finalize(metype, base_seed, skip_disabled=sim_corenrn, **kwargs)
            logging.debug("Instantiated conn %s: %d synapses", conn, syn_count)
            n_created_conns += syn_count
        return n_created_conns

    # -
    def replay(self, *_, **_kw):
        logging.warning("Replay is not available in %s", self.__class__.__name__)


# ##############
# Helper methods
# ##############

def edge_node_pop_names(edge_file, edge_pop_name, src_pop_name=None, dst_pop_name=None):
    """Find/decides the node populations names from several edge configurations

    Args:
        edge_file: The edge file to extract the population names from
        edge_pop_name: The name of the edge population
        src_pop_name: Overriding source pop name
        dst_pop_name: Overriding target population name

    Returns: tuple of the src-dst population names. Any can be None if not available
    """
    if src_pop_name and dst_pop_name or not edge_file.endswith(".h5"):
        return src_pop_name, dst_pop_name
    # Get src-dst pop names, allowing current ones to override
    src_dst_pop_names = _edge_meta_get_node_populations(edge_file, edge_pop_name) \
                        or _edge_to_node_population_names(edge_pop_name)
    if src_dst_pop_names:
        if src_pop_name is None:
            src_pop_name = src_dst_pop_names[0]
        if dst_pop_name is None:
            dst_pop_name = src_dst_pop_names[1]
    return src_pop_name, dst_pop_name


@run_only_rank0
def _edge_meta_get_node_populations(edge_file, edge_pop_name) -> Optional[tuple]:
    import h5py
    f = h5py.File(edge_file, 'r')
    if "edges" not in f:
        return None
    edge_group = f["edges"]
    if not edge_pop_name:
        assert len(edge_group) == 1, "multi-population edges require manual selection"
        edge_pop_name = next(iter(edge_group.keys()))
    edge_pop = edge_group[edge_pop_name]

    try:
        return (edge_pop["source_node_id"].attrs["node_population"],
                edge_pop["target_node_id"].attrs["node_population"])
    except KeyError:
        logging.warning("Edges don't have 'node_population' attribute")
    return None


def _edge_to_node_population_names(edge_pop_name):
    """Obtain the node source and destination population names from an edge population name
    """
    if edge_pop_name is None or "__" not in edge_pop_name:
        return None
    logging.info("(Compat) Using edge population name to know intervening nodes")
    pop_infos = edge_pop_name.split("__")
    src_pop = pop_infos[0]
    dst_pop = pop_infos[1] if len(pop_infos) > 2 else src_pop
    return src_pop, dst_pop


def _get_projection_population_id(projection):
    """Check projection config for overrides to the population ID
    """
    pop_id = None
    if "PopulationID" in projection:
        pop_id = int(projection["PopulationID"])
    if pop_id == 0:
        raise ConfigurationError("PopulationID 0 is not allowed")
    # If the projections are to be merged with base connectivity and the base
    # population is unknown, with Sonata pop we need a way to explicitly request it.
    # Note: gid offsetting must have been previously done
    if projection.get("AppendBasePopulation"):
        assert pop_id is None, "AppendBasePopulation is incompatible with PopulationID"
        log_verbose("Appending projection to base connectivity (AppendBasePopulation)")
        pop_id = 0
    return pop_id


# ######################################################################
# SynapseRuleManager
# ######################################################################
class SynapseRuleManager(ConnectionManagerBase):
    """
    The SynapseRuleManager is designed to encapsulate the creation of
    synapses for BlueBrain simulations, handling the data coming from
    the circuit file. If the BlueConfig file provides any Connection
    Rules, those override which synapses are created.

    Note that the Connection rules are processed with the assumption
    that they come in the config file from more general to more specific.
    E.g.: A column->column connection should come before
    layer 4 -> layer 2 which should come before L4PC -> L2PC.

    Once all synapses are prepared with final weights, the Netcons can be
    created.
    """

    CONNECTIONS_TYPE = SynapseReader.SYNAPSES

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Initializes a Connection/Edge manager for standard METype synapses

        Args:
            circuit_conf: The configuration object/dict
            target_manager: The (hoc-level) target manager
            cell_manager: The destination cell population manager
            src_cell_manager: The source cell population manager [Default: same as destination]
        """
        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)
        # SynapseRuleManager opens synapse file and init populations
        syn_source = circuit_conf.get("nrnPath")
        if syn_source:
            logging.info("Init %s. Options: %s", type(self).__name__, kw)
            self.open_edge_location(syn_source, circuit_conf, **kw)

    # -
    def finalize(self, base_seed=0, sim_corenrn=False, **kwargs):
        """Create the actual synapses and netcons. See super() docstring
        """
        # CoreNeuron will handle replays automatically with its own PatternStim
        kwargs.setdefault("replay_mode", ReplayMode.NONE if sim_corenrn else ReplayMode.AS_REQUIRED)
        super().finalize(base_seed, sim_corenrn, **kwargs)

    def _finalize_conns(self, tgid, conns, base_seed, sim_corenrn, **kw):
        # Note: (Compat) neurodamus hoc finalizes connections in reversed order.
        return super()._finalize_conns(tgid, conns, base_seed, sim_corenrn, reverse=True, **kw)

    # -
    def setup_delayed_connection(self, conn_config):
        """Setup delayed connection weights for synapse initialization.

        Find source and target gids and the associated connection,
        and add the delay and weight to their delay vectors.

        Args:
            conn_config: Connection configuration parsed from BlueConfig
        """
        src_target_name = conn_config["Source"]
        dst_target_name = conn_config["Destination"]
        delay = conn_config["Delay"]
        new_weight = conn_config.get("Weight", .0)

        configured_conns = 0
        for conn in self.get_target_connections(src_target_name, dst_target_name):
            conn.add_delayed_weight(delay, new_weight)
            configured_conns += 1
        return configured_conns

    # -
    @timeit(name="Replay inject")
    def replay(self, spike_manager, src_target_name, dst_target_name, start_delay=.0):
        """Create special netcons to trigger timed spikes on those synapses.

        Args:
            spike_manager: map of gids (pre-synaptic) with spike times
            src_target_name: Source population:target of the replay connections
            dst_target_name: Target whose gids should be replayed
            start_delay: Dont deliver events before t=start_delay
        """
        log_verbose("Applying replay map with %d src cells...", len(spike_manager))
        replayed_count = 0

        # Dont deliver events in the past
        if Nd.t > start_delay:
            start_delay = Nd.t
            log_verbose("Restore: Delivering events only after t=%.4f", start_delay)

        src_pop_offset = self.src_pop_offset

        for conn in self.get_target_connections(src_target_name, dst_target_name):
            raw_sgid = conn.sgid - src_pop_offset
            if raw_sgid not in spike_manager:
                continue
            conn.replay(spike_manager[raw_sgid], start_delay)
            replayed_count += 1

        total_replays = MPI.allreduce(replayed_count, MPI.SUM)
        if MPI.rank == 0:
            if total_replays == 0:
                logging.warning("No connections were injected replay stimulus")
            else:
                logging.info(" => Replaying on %d connections", total_replays)
        return total_replays



from neuron import h

pc = h.ParallelContext()
nhost = int(pc.nhost())
rank = int(pc.id())


def getNodeID(sectionID, nSeg):
    """
    Since hoc has no interface to access section node info (node_index),
    provide a python interface.
    """
    sec = h.cas()

    for seg in sec:
        h.secvec.append(sectionID)
        h.segvec.append(seg.node_index())
        nSeg += 1

    return nSeg



#!/usr/bin/env python
"""
Neuron tutorial I using the new HighLevel API. Original tutorial at
https://github.com/BlueBrain/SimulationTutorials/blob/master/CNS2017/NEURON/NEURON_intro.ipynb
"""
from neurodamus.core import Cell
from neurodamus.core import CurrentSource
from neurodamus.core import Neuron
from neurodamus.core import mechanisms


# Change v_init globally
# Alternatively v_init can be configured per simulation in run_sim(**kw)
Neuron.Simulation.v_init = -70


def test_tut1(quick=True):
    c = Cell.Builder.add_soma(60).create()
    hh = Cell.Mechanisms.HH(gkbar=0.0, gnabar=0.0, el=-70)
    hh.apply(c.soma)

    # clamp = StimuliSource.pulse(0.1, 50, delay=10).attach_to(c.soma)  # eqv. to Constant()
    CurrentSource.Constant(0.1, 50, 10).attach_to(c.soma)
    Neuron.run_sim(100, c.soma).plot()
    if quick:
        return

    # Execution with Active channels
    hh.gkbar = 0.01
    hh.gnabar = 0.2
    hh.apply(c.soma)
    sim = Neuron.run_sim(100, c.soma)
    # sim.run_continue(100)
    sim.plot()

    # Extending the model with dendrites
    c = (c.builder
         .add_dendrite("dend", 400, 9, diam=2, Ra=100)
         .add_dendrite("dend2", 400, 9, diam=2, Ra=100)
         .create())

    mechanisms.HH(el=-70, gl=5e-4, gkbar=.0, gnabar=.0).apply(c.dendrites)
    Neuron.run_sim(50, c.dendrites[0]).plot()

    Cell.show_topology()
    Neuron.h.psection(sec=c.dendrites[0])


if __name__ == "__main__":
    test_tut1(False)
    from six.moves import input
    input("Press enter to quit")



"""
Module which defines and handles Glia Cells and connectivity
"""
import libsonata
import logging
import numpy as np
import os.path

from .cell_distributor import CellDistributor
from .metype import BaseCell
from .connection import Connection
from .connection_manager import ConnectionManagerBase
from .core import EngineBase
from .core import NeurodamusCore as Nd, MPI
from .core.configuration import GlobalConfig
from .io.synapse_reader import SynapseParameters, SonataReader
from .utils import bin_search
from .utils.logging import log_verbose
from .utils.pyutils import append_recarray
from .morphio_wrapper import MorphIOWrapper


class Astrocyte(BaseCell):
    __slots__ = ('_glut_list', '_secidx2names', '_nseg_warning')

    def __init__(self, gid, meinfos, circuit_conf):
        """Instantiate a new Cell from mvd/node info."""
        super().__init__(gid, meinfos, None)
        morpho_path = circuit_conf.MorphologyPath
        morph_filename = meinfos.morph_name + "." + circuit_conf.MorphologyType
        morph_file = os.path.join(morpho_path, morph_filename)
        self._cellref, self._glut_list, self._secidx2names, self._nseg_warning \
            = self._init_cell(gid, morph_file)
        self._cellref.gid = gid

    gid = property(lambda self: int(self._cellref.gid),
                   lambda self, val: setattr(self._cellref, 'gid', val))

    endfeet = property(lambda self: self._cellref.endfeet)

    def create_endfeet(self, size):
        """
            Create endfeet sections in the cell's context.
            :param size: number of sections to create
        """
        self._cellref.execute_commands(['create endfoot[{}]'.format(size),
                                        'endfeet = new SectionList()',
                                        'forsec "endfoot" endfeet.append'])

    @staticmethod
    def _er_as_hoc(morph_wrap):
        """
            Create hoc commands for Endoplasmic Reticulum data.
            :param morph_wrap: MorphIOWrapper object holding MorphIO morphology object
        """

        '''
            For example:
                dend[0] { er_area_mcd = 0.21 er_vol_mcd = 0.4 }
                dend[1] { er_area_mcd = 0.56 er_vol_mcd = 0.23 }
                dend[2] { er_area_mcd = 1.3 er_vol_mcd = 0.78 }
                dend[3] { er_area_mcd = 0.98 er_vol_mcd = 1.1 }
        '''
        cmds = []
        cmds.extend(("{} {{ er_area_mcd = {:g} er_volume_mcd = {:g} }}".format(
            morph_wrap.section_index2name_dict[sec_index],
            er_area,
            er_vol)
            for sec_index, er_area, er_vol in zip(
            morph_wrap.morph.endoplasmic_reticulum.section_indices,
            morph_wrap.morph.endoplasmic_reticulum.surface_areas,
            morph_wrap.morph.endoplasmic_reticulum.volumes)))
        return cmds

    @staticmethod
    def _truncated_cone_surface_areas(perimeters, seg_lengths):
        radii = perimeters / (2. * np.pi)
        radii_starts = radii[:-1]
        radii_ends = radii[1:]
        slant_heights = np.sqrt(seg_lengths ** 2 + (radii_ends - radii_starts) ** 2)
        return np.pi * (radii_starts + radii_ends) * slant_heights

    @staticmethod
    def _truncated_cone_volumes(diameters, seg_lengths):
        r_begs = 0.5 * diameters[:-1]
        r_ends = 0.5 * diameters[1:]
        return (1. / 3.) * np.pi * (r_begs ** 2 + r_begs * r_ends + r_ends ** 2) * seg_lengths

    @staticmethod
    def _mcd_section_parameters(section):
        points = section.points
        diameters = section.diameters
        perimeters = section.perimeters

        segment_lengths = np.linalg.norm(points[1:] - points[:-1], axis=1)
        segment_volumes = Astrocyte._truncated_cone_volumes(diameters, segment_lengths)
        segment_surface_areas = Astrocyte._truncated_cone_surface_areas(perimeters, segment_lengths)

        total_length = segment_lengths.sum()

        # representing the entire section as a single cylinder
        perimeter = segment_surface_areas.sum() / total_length
        cross_sectional_area = segment_volumes.sum() / total_length

        return section.id, perimeter, cross_sectional_area

    @staticmethod
    def _secparams_as_hoc(morph_wrap):
        """
            Create hoc commands for section parameters (perimeters & cross-sectional area)
            :param morph_wrap: MorphIOWrapper object holding MorphIO morphology object

            For example:
                dend[0] { perimeter_mcd = 32 cross_sectional_area_mcd = 33}
        """
        cmds = []
        cmds.extend(("{} {{ perimeter_mcd = {:g} cross_sectional_area_mcd = {:g} }}".format(
            morph_wrap.section_index2name_dict[morph_sec_index + 1],
            sec_perimeter,
            sec_xsect_area)
            for morph_sec_index, sec_perimeter, sec_xsect_area in
            (Astrocyte._mcd_section_parameters(sec) for sec in morph_wrap.morph.sections)))
        return cmds

    @staticmethod
    def _init_cell(gid, morph_file):
        c = Nd.Cell(gid)
        m = MorphIOWrapper(morph_file)
        c.AddHocMorph(m.morph_as_hoc())

        glut_list = []
        c.geom_nseg_fixed()
        c.geom_nsec()  # To recount all sections

        # Insert mechanisms and populate holder lists
        logging.debug("Instantiating NGV cell gid=%d", gid)
        nseg_reduce_instance = 0  # temporary field until proper handling of nseg > 1 implemented

        for sec in c.all:
            if sec.nseg > 1:
                nseg_reduce_instance = 1
                sec.nseg = 1
            sec.insert("mcd")
            glut = Nd.GlutReceive(sec(0.5), sec=sec)
            Nd.setpointer(glut._ref_glut, 'glu2', sec(0.5).mcd)
            glut_list.append(glut)

        # Endoplasmic reticulum
        c.execute_commands(Astrocyte._er_as_hoc(m))

        # Section parameters (perimeters & cross-sectional area)
        c.execute_commands(Astrocyte._secparams_as_hoc(m))

        # Print out mcd section parameters
        # for sec in c.all:
        #     self._show_mcd()

        # Soma receiver must be last element in the glut_list
        soma = c.soma[0]
        soma.insert("mcd")
        glut = Nd.GlutReceiveSoma(soma(0.5), sec=soma)
        Nd.setpointer(glut._ref_glut, 'glu2', soma(0.5).mcd)
        glut_list.append(glut)
        return c, glut_list, m.section_index2name_dict, nseg_reduce_instance

    def _show_mcd(sec):
        if not hasattr(sec(0.5), 'mcd'):
            logging.info("No mcd mechanism found")
            return
        logging.info("{}: \tP={:.4g}\tX-Area={:.4g}\tER[area={:.4g}\tvol={:.4g}]".format(
            sec,
            sec(0.5).mcd.perimeter,
            sec(0.5).mcd.cross_sectional_area,
            sec(0.5).mcd.er_area,
            sec(0.5).mcd.er_volume)
        )

    def set_pointers(self):
        glut_list = self._glut_list
        c = self._cellref
        index = 0

        for sec in c.all:
            glut = glut_list[index]
            index += 1
            Nd.setpointer(glut._ref_glut, 'glu2', sec(0.5).mcd)
        soma = c.soma[0]
        glut = glut_list[index]
        Nd.setpointer(glut._ref_glut, 'glu2', soma(0.5).mcd)

    @property
    def glut_list(self) -> list:
        return self._glut_list

    def connect2target(self, target_pp=None):
        return Nd.NetCon(self._cellref.soma[0](1)._ref_v, target_pp,
                         sec=self._cellref.soma[0])

    @staticmethod
    def getThreshold():
        return 0.114648

    @staticmethod
    def getVersion():
        return 99


class AstrocyteManager(CellDistributor):
    # Cell Manager is the same as CellDistributor, so it's able to handle
    # the same Node formats (mvd, ...) and Cell morphologies.
    # The difference lies only in the Cell Type
    CellType = Astrocyte
    _sonata_with_extra_attrs = False

    def post_stdinit(self):
        nseg_warning = 0
        for cell in self.cells:
            cell.set_pointers()
            nseg_warning += cell._nseg_warning

        MPI.allreduce(nseg_warning, MPI.SUM)
        if nseg_warning:
            logging.warning("Astrocyte sections with multiple compartments not yet supported."
                            "Reducing %d to 1", nseg_warning)


class NeuroGliaConnParameters(SynapseParameters):
    _synapse_fields = (
        "tgid",
        "synapse_id",
        "astrocyte_section_id",
        "astrocyte_segment_id",
        "astrocyte_segment_offset"
    )


class NeuroGlialSynapseReader(SonataReader):
    LOOKUP_BY_TARGET_IDS = False
    Parameters = NeuroGliaConnParameters
    custom_parameters = set()


USE_COMPAT_SYNAPSE_ID = True
"""
Compat Synapse ID means the id is taken directly as the virtual gid, without any gap
optimization. This is still required when nrank > 1
"""


class NeuroGlialConnection(Connection):
    neurons_not_found = set()
    neurons_attached = set()

    def add_synapse(self, syn_tpoints, params_obj, syn_id=None):
        # Only store params. Glia have mechanisms pre-created
        self._synapse_params = append_recarray(self._synapse_params, params_obj)

    def finalize(self, astrocyte, base_Seed, *, base_connections=None, **kw):
        """
        Bind each glia connection to synapses in connections target cells via
        the assigned unique gid.
        """

        # TODO: Currently it receives the base_connections object to look (bin search)
        # for the sinapse to attach to. However since target cells and Glia might be
        # distributed differently across MPI ranks, this is bound to work in a single rank.
        # For the moment we fallback to using the original synapse id.

        self._netcons = []
        glut_list = astrocyte.glut_list
        n_bindings = 0
        pc = Nd.pc

        if GlobalConfig.debug_conn:
            if GlobalConfig.debug_conn == [self.tgid]:
                logging.debug("Finalizing conn %s. N params: %d", self, len(self._synapse_params))
            elif GlobalConfig.debug_conn == [self.sgid, self.tgid]:
                logging.debug("Finalizing conn %s. Params:\n%s", self, self._synapse_params)

        for syn_params in self._synapse_params:
            if USE_COMPAT_SYNAPSE_ID:
                syn_gid = 1_000_000 + syn_params.synapse_id
            else:
                tgid_conns = base_connections.get(syn_params.connected_neurons_post)
                syn_gid = self._find_neuron_endpoint_id(syn_params, tgid_conns)
                if syn_gid is None:
                    continue

            glut_obj = glut_list[int(syn_params.astrocyte_section_id)]
            netcon = pc.gid_connect(syn_gid, glut_obj)
            netcon.delay = 0.05
            self._netcons.append(netcon)

            # Soma netcon (last glut_list)
            logging.debug("[NGV] Conn %s linking synapse id %d to Astrocyte", self, syn_gid)
            netcon = pc.gid_connect(syn_gid, glut_list[-1])
            netcon.delay = 0.05
            self._netcons.append(netcon)

            n_bindings += 1
        return n_bindings

    def _find_neuron_endpoint_id(self, syn_params, conns):
        """
        Gets the endpoint id on the neuronal synapse.
        To avoid gaps, the optimized version has to search along the existing connections
        for the given synapse id.

        """
        if not conns:
            self.neurons_not_found.add(self.sgid)
            return None

        if conns[0].synapses_offset > syn_params.synapse_id:
            logging.error("Data Error: TGID %d syn offset (%d) is larger than syn gid %d",
                          conns[0].tgid, conns[0].synapses_offset, syn_params.synapse_id)
            return None

        c_i = bin_search(conns, syn_params.synapse_id, lambda c: c.synapses_offset)
        # non-exact matches are attached to the left conn (base offset)
        if len(conns) == c_i or syn_params.synapse_id < conns[c_i].synapses_offset:
            c_i -= 1
        conn = conns[c_i]
        self.neurons_attached.add(conn.tgid)

        # syn_gid: compute offset and add to the gid_base
        syn_offset = int(syn_params.synapse_id - conn.synapses_offset)
        assert syn_offset >= 0

        syn_gid = conn.syn_gid_base + syn_offset
        syn_id = conn.synapses[syn_offset].synapseID  # visible in the synapse events
        log_verbose("[GLIA ATTACH] id %d to syn Gid %d (conn %d-%d, SynID %d, syn offset %d)",
                    self.tgid, syn_gid, conn.sgid, conn.tgid, syn_id, syn_offset)
        return syn_gid


class NeuroGliaConnManager(ConnectionManagerBase):
    """A Connection Manager for Neuro-Glia connections

    NOTE: We assume the only kind of connections for Glia are Neuron-Glia
    If one day Astrocytes have connections among themselves a sub ConnectionManager
    must be used
    """

    CONNECTIONS_TYPE = "NeuroGlial"
    conn_factory = NeuroGlialConnection
    SynapseReader = NeuroGlialSynapseReader

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        kw.pop("load_offsets")
        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)

    def _add_synapses(self, cur_conn, syns_params, syn_type_restrict=None, base_id=0):
        for syn_params in syns_params:
            cur_conn.add_synapse(None, syn_params)

    def finalize(self, base_Seed=0, *_):
        """
        Instantiate connections to the simulator.

        This is a two-step process:
        First we create netcons to listen events on target synapses.Ustate,
        and assign them a virtual gid.
        Second, as part of NeuroGlialConnection.finalize(), we attach netcons to
        the target glia cell, listening for "signals" from the virtual gids.
        """
        logging.info("Creating virtual cells on target Neurons for coupling to GLIA...")
        base_manager = next(self._src_cell_manager.connection_managers.values())

        if USE_COMPAT_SYNAPSE_ID:
            total_created = self._create_synapse_ustate_endpoints(base_manager)
        else:
            total_created = self._create_synapse_ustate_endpoints_optimized(base_manager)

        logging.info("(RANK 0) Created %d Virtual GIDs for synapses.", total_created)

        super().finalize(base_Seed,
                         base_connections=base_manager.get_population(0),
                         conn_type="NeuronGlia connections")

        if not USE_COMPAT_SYNAPSE_ID:
            logging.info("Target cells coupled to: %s", NeuroGlialConnection.neurons_attached)

        if NeuroGlialConnection.neurons_not_found:
            logging.warning("Missing cells to couple Glia to: %d",
                            len(NeuroGlialConnection.neurons_not_found))

    def _create_synapse_ustate_endpoints(self, base_manager):
        """
        Creating an endpoint netcon to listen for events in synapse.Ustate
        Netcon ids are directly the synapse id (hence we are limited in number space)
        """
        pc = Nd.pc
        syn_gid_base = 1_000_000  # Below 1M is reserved for cell ids
        total_created = 0

        for conn in base_manager.all_connections():
            syn_objs = conn.synapses
            tgid_syn_offset = syn_gid_base + conn.synapses_offset
            logging.debug("Tgid: %d, Base syn offset: %d", conn.tgid, tgid_syn_offset)

            for param_i, sec in conn.sections_with_synapses:
                if conn.synapse_params[param_i].synType >= 100:  # Only Excitatory
                    synapse_gid = tgid_syn_offset + param_i
                    pc.set_gid2node(synapse_gid, MPI.rank)
                    netcon = Nd.NetCon(syn_objs[param_i]._ref_Ustate, None, 0, 0, 1.1, sec=sec)
                    pc.cell(synapse_gid, netcon)
                    conn._netcons.append(netcon)
                    total_created += 1

        return total_created

    def _create_synapse_ustate_endpoints_optimized(self, base_manager):
        # This is an optimized version to avoid using the global synapse id
        # as the virtual gid, which would strongly limit the size of the circuits.
        pc = Nd.pc
        syn_gid_base = 100_000_000  # Below 100M is reserved for cell ids

        # Get the total amount of synapses per rank and compute the base
        # synapse_gid (sum synapse count in all previous ranks)
        syn_counts = Nd.Vector(MPI.size)
        local_syn_count = sum(len(conn.synapses) for conn in base_manager.all_connections())
        MPI.allgather(local_syn_count, syn_counts)
        if MPI.rank > 0:
            syn_gid_base += syn_counts.sum(0, MPI.rank - 1)

        for conn in base_manager.all_connections():
            # Conn objects have a placeholder (syn_gid_base) for storing the id
            # for its first synapse. This enables getting to the synapse directly
            conn.syn_gid_base = syn_gid_base

            syn_objs = conn.synapses
            syn_i = None
            logging.debug("Tgid: %d, Base syn gid: %d, Base syn offset: %d",
                          conn.tgid, conn.syn_gid_base, conn.synapses_offset)

            for syn_i, (param_i, sec) in enumerate(conn.sections_with_synapses):
                if conn.synapse_params[param_i].synType >= 100:  # Only Excitatory
                    synapse_gid = syn_gid_base + syn_i
                    pc.set_gid2node(synapse_gid, MPI.rank)
                    netcon = Nd.NetCon(syn_objs[syn_i]._ref_Ustate, None, 0, 0, 1.1, sec=sec)
                    pc.cell(synapse_gid, netcon)
                    conn._netcons.append(netcon)

            # Next base is incremented number of synapses (last syn_i + 1)
            if syn_i is not None:
                syn_gid_base += syn_i + 1

        return syn_gid_base - 100_000_000  # total created


class GlioVascularManager(ConnectionManagerBase):
    CONNECTIONS_TYPE = "GlioVascular"
    InnerConnectivityCls = None  # No synapses

    def __init__(self, circuit_conf, target_manager, cell_manager, src_cell_manager=None, **kw):

        if cell_manager.circuit_target is None:
            raise Exception(
                "Circuit target is required for GlioVascular projections")
        if "Path" not in circuit_conf:
            raise Exception("Missing GlioVascular Sonata file via 'Path' configuration")

        if "VasculaturePath" not in circuit_conf:
            logging.warning("Missing Vasculature Sonata file via 'VasculaturePath' configuration")

        super().__init__(circuit_conf, target_manager, cell_manager, src_cell_manager, **kw)
        self._astro_ids = self._cell_manager.local_nodes.raw_gids()
        self._gid_offset = self._cell_manager.local_nodes.offset

    def open_edge_location(self, sonata_source, circuit_conf, **__):
        logging.info("GlioVascular sonata file %s", sonata_source)
        # sonata files can have multiple populations. In building we only use one
        # per file, hence this two lines below to access the first and only pop in
        # the file
        edge_file, *pop = sonata_source.split(":")
        storage = libsonata.EdgeStorage(edge_file)
        pop_name = pop[0] if pop else list(storage.population_names)[0]
        self._gliovascular = storage.open_population(pop_name)

        if "VasculaturePath" in circuit_conf:
            storage = libsonata.NodeStorage(circuit_conf["VasculaturePath"])
            pop_name = list(storage.population_names)[0]
            self._vasculature = storage.open_population(pop_name)

    def create_connections(self, *_, **__):
        logging.info("Creating GlioVascular virtual connections")
        # Retrieve endfeet selections for GLIA gids on the current processor
        for astro_id in self._astro_ids:
            self._connect_endfeet(astro_id)

    def _connect_endfeet(self, astro_id):

        endfeet = self._gliovascular.afferent_edges(astro_id-1)  # 0-based for libsonata API
        if endfeet.flat_size > 0:
            # Get endfeet input
            parent_section_ids = self._gliovascular.get_attribute('astrocyte_section_id', endfeet)
            lengths = self._gliovascular.get_attribute('endfoot_compartment_length', endfeet)
            diameters = self._gliovascular.get_attribute('endfoot_compartment_diameter', endfeet)
            perimeters = self._gliovascular.get_attribute('endfoot_compartment_perimeter', endfeet)

            # Retrieve instantiated astrocyte
            astrocyte = self._cell_manager.gid2cell[astro_id + self._gid_offset]

            # Create endfeet SectionList
            astrocyte.create_endfeet(parent_section_ids.size)

            # Iterate through endfeet: insert mechanisms, set values and connect to parent section
            for sec, parent_section_id, l, d, p in zip(astrocyte.endfeet,
                                                       parent_section_ids,
                                                       lengths,
                                                       diameters,
                                                       perimeters):
                sec.L = l
                sec.diam = d
                # here we just insert the mechanism. Population comes after
                sec.insert('vascouplingB')
                sec.insert('mcd')
                sec(0.5).mcd.perimeter = p
                glut = Nd.GlutReceive(sec(0.5), sec=sec)
                Nd.setpointer(glut._ref_glut, 'glu2', sec(0.5).mcd)
                # because soma glut must be the last
                astrocyte._glut_list.insert(len(astrocyte._glut_list)-1, glut)
                exec('parent_sec = astrocyte.CellRef.{}; sec.connect(parent_sec)'.format(
                    astrocyte._secidx2names[parent_section_id + 1]))
                # astrocyte.CellRef.all.append(sec)
            # Some useful debug lines:
            # cell = astrocyte.CellRef
            # logging.warn(str(cell.endfeet.printnames()))  # print endfeet section list names
            # logging.warn(str(cell.all.printnames())) #  print astrocyte names for "all" sections
            # logging.warn(str(Nd.h.topology()))  # print astrocyte topology
            # Nd.h('forall psection()')

            assert self._gliovascular.source == "vasculature"
            if hasattr(self, "_vasculature"):

                vasc_node_ids = libsonata.Selection(self._gliovascular.source_nodes(endfeet))
                assert vasc_node_ids.flat_size == len(list(astrocyte.endfeet))
                d_vessel_starts = self._vasculature.get_attribute("start_diameter", vasc_node_ids)
                d_vessel_ends = self._vasculature.get_attribute("end_diameter", vasc_node_ids)

                for sec, d_vessel_start, d_vessel_end in zip(
                        astrocyte.endfeet, d_vessel_starts, d_vessel_ends
                ):
                    # /4 is because we have an average of diameters and the output is a radius
                    sec(0.5).vascouplingB.R0pas = (d_vessel_start + d_vessel_end) / 4

    def finalize(self, *_, **__):
        pass  # No synpases/netcons


class NGVEngine(EngineBase):
    CellManagerCls = AstrocyteManager
    ConnectionTypes = {
        "NeuroGlial": NeuroGliaConnManager,
        "GlioVascular": GlioVascularManager
    }



"""
Internal module which defines an MPI object based on Neuron's ParallelContext
"""
import logging
import sys
import time
from ._neuron import Neuron


class OtherRankError(RuntimeError):
    pass


class _MPI(object):
    """A singleton of MPI runtime information
    """
    _size = 1
    _rank = 0
    _pc = None

    SUM = 1
    MAX = 2
    MIN = 3

    @classmethod
    def _init_pc(cls):
        if cls._pc is not None:
            return

        Neuron.load_hoc("netparmpi")
        cls._pc = pc = Neuron.ParallelContext()
        cls._rank = int(pc.id())
        cls._size = int(pc.nhost())

        if cls._size <= 1:
            return  # done

        # In case neurodmus is not being used as a program, but rather imported by another
        # program (API mode) we must handle exceptions instead of just killing the process
        def excepthook(etype, value, tb):
            if etype == OtherRankError:
                # Do nothing because the exception only exists because of other ranks, not locally
                return

            # Print exceptions local to this rank
            time.sleep(0.01 * cls._rank)  # Order errors
            logging.critical(str(value), exc_info=True)
            if cls._rank == 0:
                import traceback
                traceback.print_tb(tb)

            # Participate in check_no_errors allreduce, letting it know there was exception
            pc.allreduce(1, 1)  # Share error state

        sys.excepthook = excepthook

    @classmethod
    def check_no_errors(cls):
        # All processes send their status. If one is problematic it will either call MPI_Abort
        # or, in API mode, participate in the allreduce
        res = cls._pc.allreduce(0, 1)
        if res > 0:
            raise OtherRankError("Another rank raised an irrecoverable error")

    @property
    def size(self):
        self._init_pc()
        return self._size

    @property
    def rank(self):
        self._init_pc()
        return self._rank

    @property
    def pc(self):
        self._init_pc()
        return self._pc

    def __getattr__(self, name):
        if name.startswith("__"):
            return object.__getattribute__(self, name)
        self._init_pc()
        return getattr(self._pc, name)

    def py_sum(self, local_counter, aggregated_object):
        """
        An MPI function which gathers all local objects to rank 0 and sums them
        This is best suited for non-pod python objects which support sum, like `Counter`
        """
        all_counters = [local_counter] + [None] * (MPI.size - 1)  # send to rank0
        all_counters = self.pc.py_alltoall(all_counters)
        if MPI.rank == 0:
            for counter in all_counters:
                aggregated_object += counter
        return aggregated_object

    def py_reduce(self, local_counter, aggregated_object, reduce_f):
        """
        An MPI function which gathers all local objects to rank 0 and reduces them according to a
        reducing function. This is best suited for non-pod Python objects
        """
        all_objects = [local_counter] + [None] * (MPI.size - 1)  # send to rank0
        all_objects = self.pc.py_alltoall(all_objects)
        if MPI.rank == 0:
            for obj in all_objects:
                reduce_f(aggregated_object, obj)
        return aggregated_object


MPI = _MPI()
"""A singleton of MPI runtime information"""



"""
Module to load configuration from a libsonata config
"""
import json
import libsonata
import logging
import os.path


class SonataConfig:

    __slots__ = (
        "_entries",
        "_sections",
        '_config_dir',
        '_config_json',
        '_resolved_manifest',
        'circuits',
        '_circuit_networks',
        '_sim_conf',
        '_bc_circuits',
    )

    _config_entries = (
        "network", "target_simulator", "node_sets_file", "node_set"
    )
    _config_sections = (
        "run", "conditions", "output", "inputs", "reports"
    )
    # New defaults in Sonata config (not applicable to BlueConfig)
    _defaults = {
        "network": "circuit_config.json",
    }
    _path_entries_without_suffix = (
        "network",
    )

    def __init__(self, config_path):
        self._config_dir = os.path.abspath(os.path.dirname(config_path))
        self._sim_conf = libsonata.SimulationConfig.from_file(config_path)
        self._entries = {}
        self._sections = {}

        with open(config_path) as config_fh:
            self._config_json: dict = json.load(config_fh)
        self._resolved_manifest = self._build_resolver(
            self._config_json.get("manifest") or {},
            self._config_dir
        )
        for entry_name in self._config_entries:
            value = getattr(self._sim_conf, entry_name)
            self._entries[entry_name] = value
        for section_name in self._config_sections:
            section_value = self._config_json.get(section_name, {})
            self._sections[section_name] = self._resolve_section(section_value,
                                                                 self._resolved_manifest)

        self.circuits = libsonata.CircuitConfig.from_file(self.network)
        self._circuit_networks = json.loads(self.circuits.expanded_json)["networks"]
        self._bc_circuits = self._blueconfig_circuits()

    @classmethod
    def _resolve(cls, entry, name, manifest: dict):
        if not isinstance(entry, str):
            return entry  # ints, floats... no need to resolve
        if not name.lower().endswith(("_file", "_dir")) \
                and name.lower() not in cls._path_entries_without_suffix:
            return entry  # not a path
        slash_p = entry.find("/")
        if slash_p == 0:  # abs path
            return entry
        if not entry.startswith("$"):
            return os.path.normpath(os.path.join(manifest["$__CONFIG_DIR"], entry))
        # Handle variable substitution
        if slash_p > -1:
            var_name = entry[: slash_p]
            remaining = entry[slash_p:]
        else:
            var_name = entry  # just alias
            remaining = ""
        if var_name not in manifest:
            raise Exception("Cant decode path entry {}. Unknown var {}"
                            .format(entry, var_name))
        return os.path.normpath(manifest[var_name] + remaining)

    @classmethod
    def _build_resolver(cls, manifest, config_dir):
        resolved = {"$__CONFIG_DIR": config_dir}  # special entry to resolve rel paths
        for key, value in manifest.items():
            resolved[key] = cls._resolve(value, key, resolved)
        return resolved

    @classmethod
    def _resolve_section(cls, section, manifest):
        return {key: cls._resolve(val, key, manifest) for key, val in section.items()}

    _translation = {
        # Section Names
        # -------------
        "Run": "run",
        "Conditions": "conditions",
        "Projection": None,
        "StimulusInject": "inputs",
        "Connection": "connection_overrides",
        "parsedConfigures": False,

        # Section fields
        # --------------
        "run": {
            # Mandatory
            "tstop": "Duration",
            "dt": "Dt",
            "random_seed": "BaseSeed",
            # Optional
            "tstart": "Start",
            "spike_threshold": "SpikeThreshold",
            "integration_method": "SecondOrder",
            "electrodes_file": "LFPWeightsPath"
        },
        "conditions": {
            "randomize_gaba_rise_time": "randomize_Gaba_risetime"
        },
        "projection": {
        },
        "connection_overrides": {
            "target": "Destination",
            "modoverride": "ModOverride",
            "synapse_delay_override": "SynDelayOverride",
            "neuromodulation_dtc": "NeuromodDtc",
            "neuromodulation_strength": "NeuromodStrength"
        },
        "inputs": {
            "module": "Pattern",
            "input_type": "Mode",
            "random_seed": "Seed",
            "series_resistance": "RS",
            "node_set": "Target",  # for StimulusInject
            "source": "Source",  # for StimulusInject
            "sd_percent": "SDPercent",
            "amp_cv": "AmpCV"

        },
        "reports": {
            "type": "Type",
            "cells": "Target",
            "sections": "Sections",
            "scaling": "Scaling",
            "compartments": "Compartments",
            "variable_name": "ReportOn",
            "unit": "Unit",
            "dt": "Dt",
            "start_time": "StartTime",
            "end_time": "EndTime",
            "file_name": "FileName",
            "enabled": "Enabled"
        },
        "modifications": {
            "node_set": "Target"
        }
    }

    @property
    def parsedRun(self):
        parsed_run = self._translate_dict("run", self._sim_conf.run)
        self._adapt_libsonata_fields(parsed_run)
        parsed_run["CircuitPath"] = "<NONE>"  # Sonata doesnt have default circuit
        # "OutputRoot" and "SpikesFile" will be read from self._sim_conf.output
        # once libsonata resolves the manifest info
        parsed_run["OutputRoot"] = self._sim_conf.output.output_dir
        parsed_run["config_node_sets_file"] = self.circuits.node_sets_path
        parsed_run["TargetFile"] = self.circuits.node_sets_path
        parsed_run["SpikesFile"] = self._sim_conf.output.spikes_file
        parsed_run["SpikesSortOrder"] = self._sim_conf.output.spikes_sort_order.name
        parsed_run["Simulator"] = self._sim_conf.target_simulator.name
        parsed_run["TargetFile"] = self._sim_conf.node_sets_file
        parsed_run["CircuitTarget"] = self._sim_conf.node_set
        conditions = self._sections.get("conditions")
        if conditions:
            parsed_run["Celsius"] = self._sim_conf.conditions.celsius
            parsed_run["V_Init"] = self._sim_conf.conditions.v_init
            parsed_run["ExtracellularCalcium"] = self._sim_conf.conditions.extracellular_calcium
            if hasattr(self._sim_conf.conditions, "spike_location"):
                # read SpikeLocation from "conditions" with libsonata parser 0.1.17+
                # before 0.1.17 read from "run" by calling _translate_dict
                parsed_run["SpikeLocation"] = self._sim_conf.conditions.spike_location.name
        return parsed_run

    @property
    def Conditions(self):
        conditions = {}
        blacklist = (
            "Celsius",
            "VInit",
            "ExtracellularCalcium",
            "ListModificationNames",
            "SpikeLocation",
        )
        for key, value in self._translate_dict("conditions", self._sim_conf.conditions).items():
            if key in blacklist:
                continue
            if key == "Mechanisms":
                for suffix, dict_var in value.items():
                    for name, val in dict_var.items():
                        conditions[name+"_"+suffix] = val
            else:
                conditions[key] = value
        conditions["randomize_Gaba_risetime"] = str(conditions["randomize_Gaba_risetime"])
        return {"Conditions": conditions}

    def _blueconfig_circuits(self):
        """yield blue-config-style circuits"""
        node_info_to_circuit = {
            "nodes_file": "CellLibraryFile",
            "type": "PopulationType"
        }

        if "node_set" not in self._entries:
            logging.warning("Simulating all populations from all node files...")
        network = self._circuit_networks

        def make_circuit(nodes_file, node_pop_name, population_info):
            if not os.path.isabs(nodes_file):
                nodes_file = os.path.join(os.path.dirname(self.network), nodes_file)
            circuit_conf = dict(
                CircuitPath=os.path.dirname(nodes_file) or "",
                CellLibraryFile=nodes_file,
                # Use the extended ":" syntax to filter the nodeset by the related population
                CircuitTarget=node_pop_name + ":" + (self._entries.get("node_set") or ""),
                **{
                    node_info_to_circuit.get(key, key): value
                    for key, value in population_info.items()
                }
            )
            node_prop = self.circuits.node_population_properties(node_pop_name)
            circuit_conf["MorphologyPath"] = node_prop.morphologies_dir
            circuit_conf["MorphologyType"] = "h5" if node_prop.type == "astrocyte" else "swc"
            circuit_conf["METypePath"] = node_prop.biophysical_neuron_models_dir
            if node_prop.alternate_morphology_formats:
                if "neurolucida-asc" in node_prop.alternate_morphology_formats:
                    circuit_conf["MorphologyPath"] = \
                        node_prop.alternate_morphology_formats["neurolucida-asc"]
                    circuit_conf["MorphologyType"] = "asc"
                elif "h5v1" in node_prop.alternate_morphology_formats:
                    circuit_conf["MorphologyPath"] = node_prop.alternate_morphology_formats["h5v1"]
                    circuit_conf["MorphologyType"] = "h5"
            circuit_conf["Engine"] = "NGV" if node_prop.type == "astrocyte" else "METype"

            # Find inner connectivity
            # NOTE: Inner connectivity is a special kind of projection, and represents the circuit
            # default set of connections. Even though nowadays we can potentially consider
            # all connectivity as projections, under certain circuitry, like NGV, order matters and
            # therefore we keep inner connectivity to ensure they are created in the same order,
            # respecting engine precedence
            # For edges to be considered inner connectivity they must be named "default"
            for edge_config in network.get("edges") or []:
                if "nrnPath" in circuit_conf:
                    break  # Already found

                for edge_pop_name in edge_config["populations"].keys():
                    edge_storage = self.circuits.edge_population(edge_pop_name)
                    edge_type = self.circuits.edge_population_properties(edge_pop_name).type
                    inner_pop_name = "{0}__{0}__chemical".format(node_pop_name)
                    if edge_pop_name == inner_pop_name or (
                        edge_storage.source == edge_storage.target == node_pop_name
                        and edge_type == "chemical"
                        and edge_pop_name == "default"
                    ):
                        edges_file = edge_config["edges_file"]
                        if not os.path.isabs(edges_file):
                            edges_file = os.path.join(os.path.dirname(self.network), edges_file)
                        edge_pop_path = edges_file + ":" + edge_pop_name
                        circuit_conf["nrnPath"] = edge_pop_path
                        break

            circuit_conf.setdefault("nrnPath", False)
            logging.debug("Circuit config for node pop '%s': %s", node_pop_name, circuit_conf)
            return circuit_conf

        return {
            pop_name: make_circuit(node_file_info["nodes_file"], pop_name, pop_info)
            for node_file_info in network["nodes"]
            for pop_name, pop_info in node_file_info["populations"].items()
            if pop_info.get("type") != "vasculature"
        }

    # Compat with BlueConfig circuit definitions
    Circuit = property(lambda self: self._bc_circuits)

    @property
    def parsedProjections(self):
        projection_type_convert = dict(
            chemical="Synaptic",
            electrical="GapJunction",
            synapse_astrocyte="NeuroGlial",
            endfoot="GlioVascular",
            neuromodulatory="NeuroModulation"
        )
        internal_edge_pops = set(c_conf["nrnPath"] for c_conf in self._bc_circuits.values())
        projections = {}

        for edge_config in self._circuit_networks.get("edges") or []:
            for edge_pop_name, edge_pop_config in edge_config["populations"].items():
                edge_pop = self.circuits.edge_population(edge_pop_name)
                pop_type = edge_pop_config.get("type", "chemical")
                # skip unhandled synapse type
                if pop_type not in projection_type_convert:
                    logging.warning("Unhandled synapse type: " + pop_type)
                    continue

                edges_file = edge_config["edges_file"]
                if not os.path.isabs(edges_file):
                    edges_file = os.path.join(os.path.dirname(self.network), edges_file)

                # skip inner connectivity populations
                edge_pop_path = edges_file + ":" + edge_pop_name
                if edge_pop_path in internal_edge_pops:
                    continue

                projection = dict(
                    Path=edge_pop_path,
                    Source=edge_pop.source + ":",
                    Destination=edge_pop.target + ":",
                    Type=projection_type_convert.get(pop_type)
                )
                # Reverse projection direction for Astrocyte projection: from neurons to astrocytes
                if projection.get("Type") == "NeuroGlial":
                    projection["Source"], projection["Destination"] = projection["Destination"], \
                        projection["Source"]
                if projection.get("Type") == "GlioVascular":
                    for node_file_info in self._circuit_networks["nodes"]:
                        for _, pop_info in node_file_info["populations"].items():
                            if pop_info.get("type") == "vasculature":
                                projection["VasculaturePath"] = node_file_info["nodes_file"]

                proj_name = "{0}__{1.source}-{1.target}".format(edge_pop_name, edge_pop)
                projections[proj_name] = projection

        return projections

    @property
    def parsedElectrodes(self):
        return None  # No electrodes in Sonata config

    @property
    def parsedConnects(self):
        return {libsonata_conn.name: self._translate_dict("connection_overrides", libsonata_conn)
                for libsonata_conn in self._sim_conf.connection_overrides()}

    @property
    def parsedStimuli(self):
        _input_type_translation = {
            "spikes": "Current",
            "current_clamp": "Current",
            "voltage_clamp": "Voltage",
            "extracellular_stimulation": "Extracellular",
            "conductance": "Conductance"
        }

        stimuli = {}
        for name in self._sim_conf.list_input_names:
            stimulus = self._translate_dict("inputs", self._sim_conf.input(name))
            self._adapt_libsonata_fields(stimulus)
            stimulus["Pattern"] = "SEClamp" if stimulus["Pattern"] == "seclamp" \
                else snake_to_camel(stimulus["Pattern"])
            stimulus["Mode"] = _input_type_translation.get(stimulus["Mode"], stimulus["Mode"])
            stimuli[name] = stimulus
        return stimuli

    @property
    def parsedInjects(self):
        injects = {}
        # the order of stimulus injection could lead to minor difference on the results
        # so better to preserve it as in the config file
        for name in self._sections["inputs"].keys():
            inj = self._translate_dict("inputs", self._sim_conf.input(name))
            inj.setdefault("Stimulus", name)
            injects["inject"+name] = inj
        return injects

    @property
    def parsedReports(self):
        _report_type_translation = {
            "summation": "Summation",
            "synapse": "Synapse"
        }
        reports = {}
        for name in self._sim_conf.list_report_names:
            rep = self._translate_dict("reports", self._sim_conf.report(name))
            # Adapt enums and variable names read from libsonata
            self._adapt_libsonata_fields(rep)
            # Format is SONATA with sonata_config
            rep["Format"] = "SONATA"
            rep["Type"] = _report_type_translation.get(rep["Type"], rep["Type"])
            reports[name] = rep
            rep["Scaling"] = snake_to_camel(rep["Scaling"])
        return reports

    @property
    def parsedModifications(self):
        modifications = {}
        for name in self._sim_conf.conditions.list_modification_names:
            setting = self._translate_dict("modifications",
                                           self._sim_conf.conditions.modification(name))
            self._adapt_libsonata_fields(setting)
            modifications[name] = setting
        return modifications

    def _dir(self, obj):
        return [x for x in dir(obj) if not x.startswith('__') and not callable(getattr(obj, x))]

    def _adapt_libsonata_fields(self, rep):
        for key in rep:
            # Convert enums to its string representation
            if key in ("Type", "Sections", "Scaling", "Compartments", "Mode", "Pattern",
                       "SpikeLocation"):
                if not isinstance(rep[key], str):
                    rep[key] = rep[key].name
            # Convert comma separated variable names to space separated
            if key == "ReportOn":
                rep[key] = rep[key].replace(",", " ")
            # Get the int value of the enum
            elif key == "SecondOrder":
                rep[key] = int(rep[key])

    def _translate_dict(self, section_name, libsonata_obj=None) -> dict:
        item_translation = self._translation[section_name]
        result = {}
        for att in self._dir(libsonata_obj):
            key = item_translation.get(att, snake_to_camel(att))
            parsed_value = getattr(libsonata_obj, att)
            if parsed_value is not None:
                result[key] = parsed_value
        return result

    def __getattr__(self, item):
        # Immediately return native items
        if item in self._config_entries:
            return self._entries.get(item, None)
        if item in self._config_sections:
            return self._sections.get(item, {})
        # Otherwise attempt translation
        item_tr = self._translation.get(item)
        if item_tr is None:
            logging.warning("Non-native Property needs conversion: " + item)
            return {}
        return self._entries.get(item_tr) or self._sections.get(item_tr) or {}


def snake_to_camel(word):
    return ''.join(x.capitalize() or '_' for x in word.split('_'))



#!/usr/bin/env python
"""
An example on how to use CurrentSource to create a stimulus composed of different signals
"""
from neurodamus.core import CurrentSource


def test_composed_signal():
    s = CurrentSource.train(0.5, 100, 4, 100)
    s.add_noise(0, 0.1, 100)
    s.plot(ylims=(-1, 1))


if __name__ == "__main__":
    from six.moves import input
    test_composed_signal()
    input("Press enter to quit")



"""
    neurodamus.utils
    ----------------

    The neurodamus.utils package contains generic helper modules, implementing
    several new data structures, output (logging, progress,...) helpers, timers, etc.
"""
from __future__ import absolute_import
from .pyutils import *  # NOQA



"""
Main module for handling and instantiating synaptical connections
"""
from __future__ import absolute_import
import logging
import numpy as np
from os import path as ospath

from .connection_manager import ConnectionManagerBase
from .core.configuration import ConfigurationError
from .io.synapse_reader import SynapseReader, SynReaderNRN, SonataReader, SynapseParameters, \
    FormatNotSupported
from .utils import compat
from .utils.logging import log_verbose


class GapJunctionConnParameters(SynapseParameters):
    # Attribute names of synapse parameters, consistent with the normal synapses
    _synapse_fields = ("sgid", "isec", "offset", "weight", "D", "F", "ipt", "location")

    @classmethod
    def create_array(cls, length):
        npa = np.recarray(length, cls.dtype)
        npa.location = 0.5
        return npa


class GapJunctionSynapseReader(SynapseReader):
    """ Derived from SynapseReader, used for reading GapJunction synapses.
        Factory create() will attempt to instantiate GapJunctionSynToolReader,
        followed by SynReaderNRN.
    """

    @classmethod
    def create(cls, syn_src, conn_type, population=None, *args, **kw):
        """Instantiates a synapse reader, giving preference to GapJunctionSynToolReader
        """
        if fn := cls._get_sonata_circuit(syn_src):
            log_verbose("[SynReader] Using SonataReader.")
            return GapJunctionSonataReader(fn, conn_type, population, **kw)

        # Syn2 support dropped (July 2023)
        if not ospath.isdir(syn_src) and not syn_src.endswith(".h5"):
            raise FormatNotSupported("File: {syn_src}")
        logging.info("[GapJunctionSynReader] Attempting legacy hdf5 reader.")
        return SynReaderNRN(syn_src, conn_type, None, *args, **kw)


class GapJunctionSonataReader(SonataReader):
    Parameters = GapJunctionConnParameters
    parameter_mapping = {
        "weight": "conductance",
        "D": "efferent_junction_id",
        "F": "afferent_junction_id",
    }
    # "isec", "ipt", "offset" are custom parameters as in base class


class GapJunctionManager(ConnectionManagerBase):
    """
    The GapJunctionManager is similar to the SynapseRuleManager. It will
    open dedicated connectivity files which will have the locations and
    conductance strengths of gap junctions detected in the circuit.
    The user will have the capacity to scale the conductance weights.
    """

    CONNECTIONS_TYPE = SynapseReader.GAP_JUNCTIONS
    _gj_offsets = None
    SynapseReader = GapJunctionSynapseReader

    def __init__(self, gj_conf, target_manager, cell_manager, src_cell_manager=None, **kw):
        """Initialize GapJunctionManager, opening the specified GJ
        connectivity file.

        Args:
            gj_conf: The gaps junctions configuration block / dict
            target_manager: The TargetManager which will be used to query
                targets and translate locations to points
            cell_manager: The cell manager of the target population
            src_cell_manager: The cell manager of the source population
        """
        if cell_manager.circuit_target is None:
            raise ConfigurationError(
                "No circuit target. Required when initializing GapJunctionManager")
        if "Path" not in gj_conf:
            raise ConfigurationError("Missing GapJunction 'Path' configuration")

        super().__init__(gj_conf, target_manager, cell_manager, src_cell_manager, **kw)
        self._src_target_filter = target_manager.get_target(cell_manager.circuit_target,
                                                            src_cell_manager.population_name)

    def open_synapse_file(self, synapse_file, *args, **kw):
        super().open_synapse_file(synapse_file, *args, **kw)
        src_is_dir = ospath.isdir(synapse_file)
        if src_is_dir or synapse_file.endswith("nrn_gj.h5"):
            gj_dir = synapse_file if src_is_dir else ospath.dirname(synapse_file)
            self._gj_offsets = self._compute_gj_offsets(gj_dir)

    def _compute_gj_offsets(self, gj_dir):
        log_verbose("Computing gap-junction offsets from gjinfo.txt")
        gjfname = ospath.join(gj_dir, "gjinfo.txt")
        assert ospath.isfile(gjfname), "Nrn-format GapJunctions require gjinfo.txt: %s" % gj_dir
        gj_offsets = compat.Vector()
        gj_sum = 0

        for line in open(gjfname):
            gj_offsets.append(gj_sum)  # fist gid has no offset. the final total is not used
            gid, offset = map(int, line.strip().split())
            gj_sum += 2 * offset

        return gj_offsets

    def create_connections(self, *_, **_kw):
        """Gap Junctions dont use connection blocks, connect all belonging to target"""
        self.connect_all()

    def configure_connections(self, conn_conf):
        """Gap Junctions dont configure_connections"""
        pass

    def finalize(self, *_, **_kw):
        super().finalize(conn_type="Gap-Junctions")

    def _finalize_conns(self, final_tgid, conns, *_, **_kw):
        metype = self._cell_manager.get_cell(final_tgid)

        if self._gj_offsets is None:
            for conn in reversed(conns):
                conn.finalize_gap_junctions(metype, 0, 0)
        else:
            raw_tgid_0base = final_tgid - self.target_pop_offset - 1
            src_pop_offset = self.src_pop_offset
            t_gj_offset = self._gj_offsets[raw_tgid_0base]   # Old nrn_gj uses offsets
            for conn in reversed(conns):
                raw_sgid_0base = conn.sgid - src_pop_offset - 1
                conn.finalize_gap_junctions(metype, t_gj_offset, self._gj_offsets[raw_sgid_0base])
        return len(conns)



"""
Module to parse BlueConfigs
"""
import logging
from collections import defaultdict


class BlueConfigParserError(Exception):
    pass


class BlueConfig:
    _known_sections = (
        "Run", "Connection", "Projection", "Stimulus", "StimulusInject", "Report",
        "Electrode", "Modification", "NeuronConfigure", "Circuit", "Conditions",
    )

    def __init__(self, filename):
        self._sections = defaultdict(dict)
        with open(filename) as f:
            self._parse_top(f)
        if "Run" not in self._sections:
            raise BlueConfigParserError("Section 'Run' doesn't exist")
        self._sections['Run'] = self._sections['Run']['Default']

    def _parse_top(self, f):
        is_comment = False
        for line in f:
            line = line.strip()
            if not line:
                continue
            if line[0] == '#':
                is_comment = True
                continue
            if line[0] == '{' and is_comment:
                logging.debug("Skipping BlueConfig section %s", line)
                self._skip_section(f)
                continue
            is_comment = False
            header_parts = line.split()
            if len(header_parts) != 2:
                raise BlueConfigParserError("Invalid section header: " + line)
            section = header_parts[0]
            section_data = self._parse_section(f)
            if isinstance(section_data, Exception):
                raise BlueConfigParserError(
                    "Invalid data in section '{}': {}".format(line, str(section_data)))
            self._sections[section][header_parts[1]] = section_data

    @staticmethod
    def _parse_section(file_iter):
        info = {}
        next(file_iter)  # skip {
        for line in file_iter:
            line = line.strip()
            if not line or line[0] == '#':
                continue
            if '#' in line:
                line = line[:line.index('#')].rstrip()
            if line == "}":
                break
            if line == "{":
                return BlueConfigParserError("Section not closed")
            parts = line.split(maxsplit=1)
            if len(parts) < 2:
                return BlueConfigParserError(line)
            try:
                value = float(parts[1])
            except ValueError:
                value = parts[1]
            info[parts[0]] = value
        return info

    @staticmethod
    def _skip_section(file_iter):
        for line in file_iter:
            line = line.strip(" \t#")
            if line.startswith("}"):
                break
            if line.startswith("{"):
                return BlueConfigParserError(line)

    def __getattr__(self, item):
        if item not in self._known_sections:
            raise KeyError("Config key not supported: " + item)
        return self._sections[item]  # defaultdict never excepts



"""
Definitions for implementing new Engines to handle different cell types
"""

import importlib
import logging
import os
import pkgutil


class _EngineMeta(type):
    """A metaclass providing registration for new Engines
    """
    __engines = {}
    __connection_types = {}
    __instances = {}

    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        ename = name.replace("Engine", "")
        logging.info(" * Registering Engine %s (%s.%s)", ename, cls.__module__, name)
        cls.__engines.setdefault(ename, cls)
        cls.__connection_types.update(cls.ConnectionTypes)

    @property
    def engines(cls):
        return cls.__engines.values()

    @property
    def connection_types(cls):
        return cls.__connection_types

    def get(cls, name):
        """Each engine is a singleton"""
        if name is None:
            return True  # Not setting means use default
        if name in cls.__engines:
            return cls.__engines[name]
        raise RuntimeError("Engine could not be found: " + name)

    def find_plugins(cls):
        # Find/register plugins
        plugin_module = os.environ.get("NEURODAMUS_PLUGIN")
        if plugin_module:
            importlib.import_module(plugin_module)
        # Auto import correctly named modules
        for finder, name, ispkg in pkgutil.iter_modules():
            if name.startswith('neurodamus_'):
                importlib.import_module(name)
        logging.info(" => Engines Available: %s", list(cls.__engines.keys()))
        logging.info("  : Connections Types: %s", list(cls.__connection_types.keys()))


class EngineBase(metaclass=_EngineMeta):
    """A base class to define an engine.

    Engines are the fundamental blocks to handle different kinds of cells, like
    Glia, in a plugin-like interface.
    Engines must either implement their own create_cells and create_synapses
    methods (for full flexibility) or specify which are the Manager classes.

    Specifying the Manager classes is suitable for most cases.
    Without any customization an engine will default to use
      CellManagerCls = None   # Use existing instance of CellDistributor
      SynapseManagerCls = None  # Use existing instance of SynapseRuleManager

    Such setup is equivalent to not specifying the Engine for a given circuit,
    effectively handling additional circuits by the built-in Engine.
    Specifying CellManagerCls will instantiate cells with the new Engine. If a
    SynapseManagerCls is not provided then only cell creation happens.
    Specifying SynapseManagerCls alone is not supported.

    """

    CellManagerCls = None
    InnerConnectivityCls = None
    ConnectionTypes = {}
    """A dict of the new connection types and associated Manager class"""
    CircuitPrecedence = 1
    """Precedence influences instantiation order. The lower the earlier setup"""

    @classmethod
    def new_cell_manager(cls, circuit_conf, target_parser, run_conf):
        if not cls.CellManagerCls:
            return NotImplemented
        return cls.CellManagerCls(circuit_conf, target_parser, run_conf)

    def new_synapse_manager(cls, circuit_conf, target_manager, cell_manager):
        """Routine responsible for creating connection managers.
        """
        if not cls.SynapseManagerCls:
            return NotImplemented
        if cell_manager is None:
            raise RuntimeError("Circuit Not Initialized. Please load_nodes() first")
        return cls.SynapseManagerCls(circuit_conf, target_manager, cell_manager)

    def report(report, target_parser, rep_target, corenrn_conf, cell_manager):
        """Create reports for this type of engine.
        """
        if cell_manager is None:
            raise RuntimeError("Circuit Not Initialized. Please load_nodes() first")
        return NotImplemented



# Small script to measure memory usage of synapse objects in NEURON.
# Usage: nrniv (or special) -python synstats.py

from neuron import h
import os


def get_mem_usage():
    """
    Return memory usage information in KB.
    """
    with open("/proc/self/statm") as fd:
        _, data_size, _ = fd.read().split(maxsplit=2)
    usage_kb = float(data_size) * os.sysconf("SC_PAGE_SIZE") / 1024

    return usage_kb


# Dummy class to pass to synapse objects
class SynParams:
    def __getattr__(self, item):
        return 1


n_inst = 1000000

h.load_file("RNGSettings.hoc")
h.load_file("Map.hoc")

map_hoc = h.Map()
RNGset = h.RNGSettings()
RNGset.interpret(map_hoc)

pc = h.ParallelContext()

sec = h.Section()
sec.push()
params_obj = SynParams()

h.load_file("AMPANMDAHelper.hoc")
mem = get_mem_usage()
AMPA_helper = [h.AMPANMDAHelper(1, params_obj, 0.5, i, 0) for i in range(n_inst)]
netcon_ampa = [pc.gid_connect(1000, helper.synapse) for helper in AMPA_helper]
mem2 = get_mem_usage()
print('Memory usage per object ProbAMPA: %f KB' % ((mem2 - mem) / n_inst))

h.load_file("GABAABHelper.hoc")
mem = get_mem_usage()
GABAAB_helper = [h.GABAABHelper(1, params_obj, 0.5, i, 0) for i in range(n_inst)]
netcon_gabaab = [pc.gid_connect(1000, helper.synapse) for helper in GABAAB_helper]
mem2 = get_mem_usage()
print('Memory usage per object ProbGABAAB: %f KB' % ((mem2 - mem) / n_inst))

h.load_file("GluSynapseHelper.hoc")
mem = get_mem_usage()
GluSynapse_helper = [h.GluSynapseHelper(1, params_obj, 0.5, i, 0, map_hoc) for i in range(n_inst)]
netcon_glu = [pc.gid_connect(1000, helper.synapse) for helper in GluSynapse_helper]
mem2 = get_mem_usage()
print('Memory usage per object GluSynapse: %f KB' % ((mem2 - mem) / n_inst))

mem = get_mem_usage()
Gap_helper = [h.Gap(0.5) for i in range(n_inst)]
mem2 = get_mem_usage()
print('Memory usage per object Gap: %f KB' % ((mem2 - mem) / n_inst))



"""
Collection of core helpers / utilities
"""
from __future__ import absolute_import
import time
from array import array
from functools import wraps
from inspect import Signature, signature
from ._mpi import MPI
from ..utils import progressbar


class ProgressBarRank0(progressbar.Progress):
    """Helper Progressbar that only shows on Rank 0.
       For MPI clusters size > 1 it always uses simplified bars. Otherwise auto-detects (isatty).
    """
    def __new__(cls, end, *args, **kwargs):
        if MPI.rank == 0:
            return progressbar.ProgressBar(end, *args, tty_bar=(MPI.size == 1) and None, **kwargs)
        return progressbar.Progress(end, *args, **kwargs)


def mpi_no_errors(f):
    """Convenience decorator which checks all processes are fine when f returns
    """
    @wraps(f)
    def mpi_ok_wrapper(*args, **kw):
        # Three scenarios:
        #   0 - all ranks return normally: they all wait in the barrier
        #   1 - all ranks throw exception: exception propagated up, no barrier hit
        #   2 - some ranks throw exception. Good ranks will wait in barrier. Two options:
        #       In program mode bad ranks must call MPI_abort (in commands.py)
        #       In API mode we must catch unhandled exception participate in collective count
        #       and raise exception in all ranks
        res = f(*args, **kw)
        if MPI.size > 1:
            MPI.check_no_errors()
        return res

    return mpi_ok_wrapper


class run_only_rank0:
    """Decorator that makes a given func to run only in rank 0.

    It will broadcast results IFF the user specifies return type notation.
    It handles nested level to avoid broadcasting while we are already in rank0_only mode
    """
    nested_depth = 0

    def __new__(cls, f):
        has_return = signature(f).return_annotation != Signature.empty

        @wraps(f)
        def rank0_wrapper(*args, **kw):
            # Situation we dont need/want the broadcast
            if MPI.size == 1 or cls.nested_depth > 0:
                return f(*args, **kw)

            cls.nested_depth += 1
            res = f(*args, **kw) if MPI.rank == 0 else None
            cls.nested_depth -= 1

            if has_return:
                return MPI.py_broadcast(res, 0)

        return rank0_wrapper


def return_neuron_timings(f):
    """Decorator to collect and return timings on a neuron run
    """
    @wraps(f)
    def timings_wrapper(*args, **kw):
        # Timings structure (being returned)
        tdat = array("d", [.0]*8)
        tstart = time.time()
        pc = MPI.pc
        wait_base = pc.wait_time()

        f(*args, **kw)  # Discard return values

        tdat[0] = pc.wait_time() - wait_base
        tdat[1] = pc.step_time()
        tdat[2] = pc.send_time()
        tdat[3] = pc.vtransfer_time()
        tdat[4] = pc.vtransfer_time(1)  # split exchange time
        tdat[6] = pc.vtransfer_time(2)  # reduced tree computation time
        tdat[4] -= tdat[6]
        tdat[7] = time.time() - tstart      # total time
        return tdat

    return timings_wrapper



"""
Mechanisms to load and balance cells across the computing resources.
"""
from __future__ import absolute_import, print_function
import abc
import hashlib
import logging  # active only in rank 0 (init)
import os
import weakref
from contextlib import contextmanager
from enum import Enum
from io import StringIO
from os import path as ospath
from pathlib import Path

import numpy

from .connection_manager import ConnectionManagerBase
from .core import MPI, mpi_no_errors, run_only_rank0
from .core import NeurodamusCore as Nd
from .core import ProgressBarRank0 as ProgressBar
from .core.configuration import ConfigurationError
from .core.configuration import GlobalConfig, LoadBalanceMode, LogLevel, find_input_file, SimConfig
from .core.nodeset import NodeSet
from .io import cell_readers
from .lfp_manager import LFPManager
from .metype import Cell_V5, Cell_V6, EmptyCell
from .target_manager import TargetSpec
from .utils import compat
from .utils.logging import log_verbose, log_all
from .utils.memory import DryRunStats, get_mem_usage_kb


class NodeFormat(Enum):
    NCS = 1
    MVD3 = 2
    SONATA = 3


class VirtualCellPopulation:
    """
    A virtual cell population offers a compatible interface with Cell Manager,
    however it doesnt instantiate cells.
    It is mostly used as source of projections
    """
    _total_count = 0

    def __init__(self, population_name, gids=None, circuit_target=None):
        """Initializes a VirtualCellPopulation

        A virtual manager will have minimal set of attributes, namely
        the population name, the target name and the NodeSet
        """
        self.population_name = population_name
        self.circuit_target = circuit_target
        self.local_nodes = NodeSet(gids).register_global(population_name or '')
        VirtualCellPopulation._total_count += 1
        if VirtualCellPopulation._total_count > 1:
            logging.warning("For non-sonata circuit, "
                            "only a single Virtual Cell Population works with REPLAY")

    is_default = property(lambda self: False)
    is_virtual = property(lambda self: True)

    def __str__(self):
        return "([VIRT] {:s})".format(self.population_name)


class _CellManager(abc.ABC):

    @abc.abstractmethod
    def get_final_gids(self):
        ...

    @abc.abstractmethod
    def get_cell(self, gid):
        ...

    def get_cellref(self, gid):
        """Retrieve a cell object given its gid.
        Note that this function handles multisplit cases incl converting to an
        spgid automatically \n
        Returns: Cell object
        """
        if self._binfo:
            # are we in load balance mode? must replace gid with spgid
            gid = self._binfo.thishost_gid(gid)
        return self._pc.gid2obj(gid)

    # Methods for compat with hoc

    @abc.abstractmethod
    def getGidListForProcessor(self):
        ...

    def getMEType(self, gid):
        return self.get_cell(gid)

    def getCell(self, gid):
        return self.get_cellref(gid)


class CellManagerBase(_CellManager):

    CellType = NotImplemented  # please override
    """The underlying Cell type class

    signature:
        __init__(self, gid, cell_info, circuit_conf)
    """

    _node_loader = None
    """Default function implementing the loading of nodes data, a.k.a. MVD

    signature:
        load(circuit_conf, gidvec, stride=1, stride_offset=0)
    """

    _node_format = NodeFormat.SONATA  # NCS, Mvd, Sonata...
    """Default Node file format"""

    def __init__(self, circuit_conf, target_manager, _run_conf=None, **_kw):
        """Initializes CellDistributor

        Args:
            circuit_conf: The "Circuit" blueconfig block
            target_parser: the target parser hoc object, to retrieve target cells' info
            run_conf: Run configuration

        """
        self._circuit_conf = circuit_conf
        self._circuit_name = circuit_conf._name
        self._target_manager = target_manager
        self._target_spec = TargetSpec(circuit_conf.CircuitTarget)
        self._population_name = None
        self._local_nodes = None
        self._total_cells = 0   # total cells in target, being simulated
        self._gid2cell = {}

        self._global_seed = 0
        self._ionchannel_seed = 0
        self._binfo = None
        self._pc = Nd.pc
        self._conn_managers_per_src_pop = weakref.WeakValueDictionary()

        if type(circuit_conf.CircuitPath) is str:
            self._init_config(circuit_conf, self._target_spec.population or '')
        else:
            logging.info(" => %s Circuit has been disabled", self.circuit_name or "(default)")

    # read-only properties
    target_manager = property(lambda self: self._target_manager)
    local_nodes = property(
        lambda self: self._local_nodes if self._local_nodes is not None else NodeSet()
    )
    total_cells = property(lambda self: self._total_cells)
    cells = property(lambda self: self._gid2cell.values())
    gid2cell = property(lambda self: self._gid2cell)
    pc = property(lambda self: self._pc)
    population_name = property(lambda self: self._population_name)
    circuit_target = property(lambda self: self._target_spec.name)
    circuit_name = property(lambda self: self._circuit_name)
    is_default = property(lambda self: self._circuit_name is None)
    is_virtual = property(lambda self: False)
    connection_managers = property(lambda self: self._conn_managers_per_src_pop)

    def is_initialized(self):
        return self._local_nodes is not None

    def __str__(self):
        return "({}: {})".format(self.__class__.__name__, str(self._population_name))

    # Compatibility with neurodamus-core (used by TargetManager, CompMapping)
    # Create hoc vector from numpy.array
    def getGidListForProcessor(self):
        return compat.hoc_vector(self.local_nodes.final_gids())

    def get_final_gids(self):
        return numpy.array(self.local_nodes.final_gids())

    def _init_config(self, circuit_conf, pop):
        if self._node_format == NodeFormat.SONATA:
            if not ospath.isabs(circuit_conf.CellLibraryFile):
                circuit_conf.CellLibraryFile = find_input_file(circuit_conf.CellLibraryFile)
            if not pop:   # Last attempt to get pop name
                pop = self._get_sonata_population_name(circuit_conf.CellLibraryFile)
                logging.info(" -> Discovered node population name: %s", pop)
        if not pop and circuit_conf._name:
            pop = circuit_conf._name
            logging.warning("(Compat) Assuming population name from Circuit: %s", pop)
        self._population_name = pop
        if not pop:
            logging.warning("Could not discover population name. Assuming '' (empty)")
            if not self.is_default:
                raise Exception("Only the default population can be unnamed")
        is_base_pop = self.is_default or circuit_conf.get("no_offset")
        self._local_nodes = NodeSet().register_global(pop, is_base_pop)

    @classmethod
    def _get_sonata_population_name(self, node_file):
        import libsonata  # only for SONATA
        pop_names = libsonata.NodeStorage(node_file).population_names
        assert len(pop_names) == 1
        return next(iter(pop_names), None)

    def load_nodes(self, load_balancer=None, *, _loader=None, loader_opts=None):
        """Top-level loader of nodes.
        """
        if self._local_nodes is None:
            return
        conf = self._circuit_conf
        _loader = _loader or self._node_loader
        loader_f = (lambda *args: _loader(*args, **loader_opts)) if loader_opts else _loader

        logging.info("Reading Nodes (METype) info from '%s'", conf.CellLibraryFile)
        if load_balancer and load_balancer.population != self._target_spec.population:
            log_verbose("Load balance object doesn't apply to '%s'", self._target_spec.population)
            load_balancer = None
        if not load_balancer or SimConfig.dry_run:
            gidvec, me_infos, *cell_counts = self._load_nodes(loader_f)
        else:
            gidvec, me_infos, *cell_counts = self._load_nodes_balance(loader_f, load_balancer)
        self._local_nodes.add_gids(gidvec, me_infos)
        self._total_cells = cell_counts[0]
        logging.info(" => Loaded info about %d target cells (out of %d)", *cell_counts)

    def _load_nodes(self, loader_f):
        """Base loader which handles targets"""
        conf = self._circuit_conf
        target_spec = self._target_spec

        if target_spec.name:
            logging.info(" -> Distributing '%s' target cells Round-Robin", target_spec)
            target_gids = self._target_manager.get_target(target_spec).get_raw_gids()
            gidvec, me_infos, full_size = loader_f(conf, target_gids, MPI.size, MPI.rank)
            total_cells = len(target_gids)
        else:
            logging.info(" -> Distributing ALL cells Round-Robin (No target provided)")
            gidvec, me_infos, full_size = loader_f(conf, None, MPI.size, MPI.rank)
            total_cells = full_size
        return gidvec, me_infos, total_cells, full_size

    def _load_nodes_balance(self, loader_f, load_balancer):
        target_spec = self._target_spec
        if not load_balancer.valid_load_distribution(target_spec):
            raise RuntimeError("No valid Load Balance info could be found or derived."
                               "Please perform a full load balance.")

        logging.info(" -> Distributing target '%s' using Load-Balance", target_spec.name)
        self._binfo = load_balancer.load_balance_info(target_spec)
        # self._binfo has gidlist, but gids can appear multiple times
        all_gids = numpy.unique(
            self._binfo.gids.as_numpy().astype("uint32") - self._local_nodes.offset
        )
        total_cells = len(all_gids)
        gidvec, me_infos, full_size = loader_f(self._circuit_conf, all_gids)
        return gidvec, me_infos, total_cells, full_size

    # -
    def finalize(self, **opts):
        """Instantiates cells and initializes the network in the simulator.

        Note: it should be called after all cell distributors have done load_nodes()
            so gids offsets are final.
        """
        if self._local_nodes is None:
            return
        logging.info("Finalizing cells... Gid offset: %d", self._local_nodes.offset)
        self._instantiate_cells(**opts)
        self._update_targets_local_gids()
        self._init_cell_network()
        self._local_nodes.clear_cell_info()

    @mpi_no_errors
    def _instantiate_cells(self, _CellType=None, **_opts):
        CellType = _CellType or self.CellType
        assert CellType is not None, "Undefined CellType in Manager"
        Nd.execute("xopen_broadcast_ = 0")

        logging.info(" > Instantiating cells... (%d in Rank 0)", len(self._local_nodes))
        cell_offset = self._local_nodes.offset

        if GlobalConfig.verbosity >= LogLevel.DEBUG:
            gid_info_items = self._local_nodes.items()
        else:
            gid_info_items = ProgressBar.iter(self._local_nodes.items(), len(self._local_nodes))

        for gid, cell_info in gid_info_items:
            cell = CellType(gid, cell_info, self._circuit_conf)
            self._store_cell(gid + cell_offset, cell)

    @mpi_no_errors
    def _instantiate_cells_dry(self, CellType, skip_metypes, **_opts):
        """
        Instantiates the subset of selected cells while measuring memory taken by each metype

        Args:
            CellType: The cell type class
            full_memory_counter: The memory counter to be updated for each metype
        """
        assert CellType is not None, "Undefined CellType in Manager"
        Nd.execute("xopen_broadcast_ = 0")

        logging.info(" > Dry run on cells... (%d in Rank 0)", len(self._local_nodes))
        cell_offset = self._local_nodes.offset
        gid_info_items = self._local_nodes.items()

        prev_metype = None
        prev_memory = get_mem_usage_kb()
        metype_n_cells = 0
        memory_dict = {}
        MAX_CELLS = 50

        def store_metype_stats(metype, n_cells):
            nonlocal prev_memory
            end_memory = get_mem_usage_kb()
            memory_allocated = end_memory - prev_memory
            log_all(logging.DEBUG, " * METype %s: %.1f KiB averaged over %d cells",
                    metype, memory_allocated/n_cells, n_cells)
            memory_dict[metype] = max(0, memory_allocated / n_cells)
            prev_memory = end_memory

        for gid, cell_info in gid_info_items:
            if cell_info is None:
                continue
            metype = "{0.mtype}-{0.etype}".format(cell_info)
            if metype in skip_metypes:
                continue
            if prev_metype is not None and metype != prev_metype:
                store_metype_stats(prev_metype, metype_n_cells)
                metype_n_cells = 0
            if metype_n_cells >= MAX_CELLS:
                continue
            cell = CellType(gid, cell_info, self._circuit_conf)
            self._store_cell(gid + cell_offset, cell)
            prev_metype = metype
            metype_n_cells += 1

        if prev_metype is not None and metype_n_cells > 0:
            store_metype_stats(prev_metype, metype_n_cells)

        return memory_dict

    def _update_targets_local_gids(self):
        logging.info(" > Updating targets")
        cell_offset = self._local_nodes.offset
        if cell_offset and self._target_spec.name:
            target = self._target_manager.get_target(self._target_spec)
            if not hasattr(target, "set_offset"):
                raise NotImplementedError("No gid offsetting supported by neurodamus Target.hoc")
            target.set_offset(cell_offset)
        # Add local gids to matching targets
        self._target_manager.register_local_nodes(self._local_nodes)

    def _init_cell_network(self):
        """Init global gids for cell networking
        """
        logging.info(" > Initializing cell network")
        self._init_rng()
        pc = self._pc

        for final_gid, cell in self._gid2cell.items():
            cell.re_init_rng(self._ionchannel_seed)
            nc = cell.connect2target(None)  # Netcon doesnt require being stored

            if self._binfo:
                gid_i = int(self._binfo.gids.indwhere("==", final_gid))
                cb = self._binfo.bilist.object(self._binfo.cbindex.x[gid_i])
                # multisplit cells call cb.multisplit() instead
                if cb.subtrees.count() > 0:
                    cb.multisplit(nc, self._binfo.msgid, pc, pc.id())
                    cell.gid = final_gid
                    continue

            pc.set_gid2node(final_gid, pc.id())
            pc.cell(final_gid, nc)
            cell.gid = final_gid  # update the cell.gid last (RNGs had to use the base gid)

        pc.multisplit()

    def enable_report(self, report_conf, target_name, use_coreneuron):
        """Placeholder for Engines implementing their own reporting

        Args:
            report_conf: The dict containing the report configuration
            target_name: The target of the report
            use_coreneuron: Whether the simulator is CoreNeuron
        """
        pass

    def load_artificial_cell(self, gid, artificial_cell):
        logging.info(" > Adding Artificial cell for CoreNeuron")
        cell = EmptyCell(gid, artificial_cell)
        self._store_cell(gid, cell)
        nc = cell.connect2target(None)
        self._pc.set_gid2node(cell.gid, self._pc.id())
        self._pc.cell(cell.gid, nc)

    def _init_rng(self):
        rng_info = Nd.RNGSettings()
        self._global_seed = rng_info.getGlobalSeed()
        self._ionchannel_seed = rng_info.getIonChannelSeed()
        return rng_info

    def _store_cell(self, gid, cell):
        self._gid2cell[gid] = cell

    def get_cell(self, gid):
        return self._gid2cell[gid]

    def get_cellref(self, gid):
        return self._gid2cell[gid]._cellref

    def record_spikes(self, gids=None, append_spike_vecs=None):
        """Setup recording of spike events (crossing of threshold) for cells on this node
        """
        if not self._local_nodes:
            return
        spikevec, idvec = append_spike_vecs or (Nd.Vector(), Nd.Vector())
        if gids is None:
            gids = self._local_nodes.final_gids()

        for gid in gids:
            # only want to collect spikes of cell pieces with the soma (i.e. the real gid)
            if not self._binfo or self._binfo.thishost_gid(gid) == gid:
                self._pc.spike_record(gid, spikevec, idvec)
        return spikevec, idvec

    def register_connection_manager(self, conn_manager: ConnectionManagerBase):
        src_population = conn_manager.src_cell_manager.population_name
        if src_population in self._conn_managers_per_src_pop:
            logging.warning("Skip registering %s as a second pop source", conn_manager)
        else:
            self._conn_managers_per_src_pop[src_population] = conn_manager

    def post_stdinit(self):
        """Post stdinit actions"""
        if self._circuit_conf.DetailedAxon:
            for c in self.cells:
                c.delete_axon()


class GlobalCellManager(_CellManager):
    """
    GlobalCellManager is a wrapper over all Cell Managers so that we can query
    any cell from its global gid
    """

    def __init__(self):
        self._cell_managers = []
        self._pc = Nd.pc
        self._lfp_manager = LFPManager()

    def register_manager(self, cell_manager):
        self._cell_managers.append(cell_manager)

    def finalize(self):
        self._cell_managers.sort(key=lambda x: x.local_nodes.offset)

    # Accessor methods (Keep CamelCase API for compatibility with existing hoc)
    # ----------------
    def getGidListForProcessor(self):

        def _hoc_append(vec_a, vec_b):
            return vec_a.append(vec_b)

        from functools import reduce
        return reduce(_hoc_append, (man.getGidListForProcessor() for man in self._cell_managers))

    def get_final_gids(self):
        return numpy.concatenate([man.get_final_gids() for man in self._cell_managers])

    def _find_manager(self, gid):
        cell_managers_iter = iter(self._cell_managers)
        prev_manager = next(cell_managers_iter)  # base cell manager
        for manager in cell_managers_iter:
            if manager.local_nodes.offset > gid:
                break
            prev_manager = manager
        return prev_manager

    def get_cell(self, gid):
        manager = self._find_manager(gid)
        return manager.get_cell(gid)

    def get_cellref(self, gid):
        """Retrieve a cell object given its gid.
        Note that this function handles multisplit cases incl converting to an
        spgid automatically \n
        Returns: Cell object
        """
        manager = self._find_manager(gid)
        if manager._binfo:
            # are we in load balance mode? must replace gid with spgid
            gid = manager._binfo.thishost_gid(gid)
        return self._pc.gid2obj(gid)

    def getSpGid(self, gid):
        """Retrieve the spgid from a gid (provided we are using loadbalancing)

        Args:
            gid: The base gid (as read from start.ncs)

        Returns: The gid as it appears on this cpu (if this is the same as the base gid,
        then that is the soma piece)
        """
        manager = self._find_manager(gid)
        if manager._binfo:
            return manager._binfo.thishost_gid(gid)
        return gid

    def getPopulationInfo(self, gid):
        manager = self._find_manager(gid)
        return manager.population_name, manager.local_nodes.offset


class CellDistributor(CellManagerBase):
    """ Manages a group of cells for BBP simulations, V5 and V6

        Instantiated cells are stored locally (.cells property)
    """

    _cell_loaders = {
        "start.ncs": cell_readers.load_ncs,
        "circuit.mvd3": cell_readers.load_mvd3,
    }

    _sonata_with_extra_attrs = True  # Enable search extra node attributes

    def _init_config(self, circuit_conf, _pop):
        if not circuit_conf.CellLibraryFile:
            logging.warning("CellLibraryFile not set. Assuming legacy 'start.ncs'")
            circuit_conf.CellLibraryFile = "start.ncs"
        if circuit_conf.CellLibraryFile.endswith(".ncs"):
            self._node_format = NodeFormat.NCS
        elif circuit_conf.CellLibraryFile.endswith(".mvd3"):
            self._node_format = NodeFormat.MVD3

        self._is_v5_circuit = circuit_conf.CellLibraryFile == "start.ncs" or (
            circuit_conf.nrnPath and ospath.isfile(ospath.join(circuit_conf.nrnPath, "start.ncs"))
            and not ospath.isfile(ospath.join(circuit_conf.CircuitPath, "circuit.mvd3"))
        )
        if self._is_v5_circuit and self._circuit_conf.DetailedAxon:
            raise ConfigurationError("V5 circuits don't support keeping detailed axons")

        super()._init_config(circuit_conf, _pop)

    def load_nodes(self, load_balancer=None, **kw):
        """gets gids from target, splits and returns a GidSet with all metadata
        """
        loader_opts = kw.pop("loader_opts", {}).copy()
        all_cell_requirements = SimConfig.cell_requirements
        cell_requirements = all_cell_requirements.get(self._population_name) or (
            self.is_default and all_cell_requirements.get(None)
        )
        if self._node_format == NodeFormat.SONATA:
            loader = cell_readers.load_sonata
            loader_opts["node_population"] = self._population_name  # mandatory in Sonata
            loader_opts["load_dynamic_props"] = cell_requirements
            loader_opts["has_extra_data"] = self._sonata_with_extra_attrs
        else:
            if cell_requirements:
                raise Exception('Additional cell properties only available with SONATA')
            nodes_filename = self._circuit_conf.CellLibraryFile
            loader = self._cell_loaders.get(nodes_filename, cell_readers.load_mvd3)
            loader_opts = {}

        log_verbose("Nodes Format: %s, Loader: %s", self._node_format, loader.__name__)
        return super().load_nodes(load_balancer, _loader=loader, loader_opts=loader_opts)

    def _instantiate_cells(self, dry_run_stats_obj: DryRunStats = None, **opts):
        """
        Instantiates cells, honouring dry_run if provided

        Args:
            dry_run_store_stats: Do a dry-run and update the inner fields accordingly
        """
        if self.CellType is not NotImplemented:
            return super()._instantiate_cells(self.CellType)
        conf = self._circuit_conf
        CellType = Cell_V5 if self._is_v5_circuit else Cell_V6
        if conf.MorphologyType:
            CellType.morpho_extension = conf.MorphologyType

        log_verbose("Loading metypes from: %s", conf.METypePath)
        log_verbose("Loading '%s' morphologies from: %s",
                    CellType.morpho_extension, conf.MorphologyPath)
        if dry_run_stats_obj is None:
            super()._instantiate_cells(CellType, **opts)
        else:
            cur_metypes_mem = dry_run_stats_obj.metype_memory
            memory_dict = self._instantiate_cells_dry(CellType, cur_metypes_mem, **opts)
            log_verbose("Updating global dry-run memory counters with %d items", len(memory_dict))
            cur_metypes_mem.update(memory_dict)


class LoadBalance:
    """
    Class handling the several types of load_balance info, including
    generating and loading the various files.

    LoadBalance instances target the current system (cpu count) and circuit
    BUT check/create load distribution for any given target.
    The circuit is identified by the nodes file AND population.

    NOTE: Given the heavy costs of computing load balance, some state files are created
    which allow the balance info to be reused. These are

     - cx_{TARGET}.dat: File with complexity information for the cells of a given target
     - cx_{TARGET}.{CPU_COUNT}.dat: The file assigning cells/pieces to individual CPUs ranks.

    For more information refer to the developer documentation.
    """
    _base_output_dir = "sim_conf"
    _circuit_lb_dir_tpl = "_loadbal_%s.%s"      # Placeholders are (file_src_hash, population)
    _cx_filename_tpl = "cx_%s#.dat"             # use # to well delimiter the target name
    _cpu_assign_filename_tpl = "cx_%s#.%s.dat"  # prefix must be same (imposed by Neuron)

    def __init__(self, balance_mode, nodes_path, pop, target_manager, target_cpu_count=None):
        """
        Creates a new Load Balance object, associated with a given node file
        """
        self.lb_mode = balance_mode
        self.target_cpu_count = target_cpu_count or MPI.size
        self._target_manager = target_manager
        self._valid_loadbalance = set()
        self.population = pop or ""
        self._lb_dir, self._cx_targets = self._get_circuit_loadbal_dir(nodes_path, self.population)
        log_verbose("Found existing targets with loadbal: %s", self._cx_targets)

    @classmethod
    @run_only_rank0
    def _get_circuit_loadbal_dir(cls, node_file, pop) -> tuple:
        """Ensure lbal dir exists. dir may be crated on rank 0"""
        lb_dir = cls._loadbal_dir(node_file, pop)
        if lb_dir.is_dir():
            return lb_dir, cls._get_lbdir_targets(lb_dir)

        logging.info("Creating load balance dir %s", lb_dir)
        os.makedirs(lb_dir)
        return lb_dir, set()

    @classmethod
    def _get_lbdir_targets(cls, lb_dir: Path) -> list:
        """Inspects the load-balance folder and detects which targets are load balanced"""
        prefix, suffix = cls._cx_filename_tpl.split("%s")
        return set(
            fname.name[len(prefix):-len(suffix)]
            for fname in lb_dir.glob(cls._cx_filename_tpl.replace("%s", "*"))
        )

    @run_only_rank0
    def valid_load_distribution(self, target_spec: TargetSpec) -> bool:
        """Checks whether we have valid load-balance files, attempting to
        derive from larger target distributions if possible.
        """
        if (target_spec.population or "") != self.population:
            logging.info(" => Load balance Population mismatch. Requested: %s, Existing: %s",
                         target_spec.population, self.population)
            return False

        target_name = target_spec.simple_name

        # Check cache
        if target_name in self._valid_loadbalance:
            logging.info(" => Cell distribution from Load Balance is valid")
            return True

        # A valid load-distribution requires BOTH the cx and the cpu_assign files.
        if self._cx_valid(target_spec):
            cpu_assign_filename = self._cpu_assign_filename(target_name)
            if cpu_assign_filename.is_file():
                logging.info(" => Found valid load balance: %s", cpu_assign_filename)
            else:
                # Still need to derive the CPU assignment
                logging.info(" => Found valid complexity file: %s", self._cx_filename(target_name))
                self._cpu_assign(target_name)
            self._valid_loadbalance.add(target_name)
            return True

        # If not found, attempt to find a cx file computed for a superset target
        # so we can derive a new cx file without instantiating cells
        if self._reuse_cell_complexity(target_spec):
            self._cpu_assign(target_name)
            self._valid_loadbalance.add(target_name)
            return True
        return False

    # -
    def _reuse_cell_complexity(self, target_spec: TargetSpec) -> bool:
        """Check if the complexities of all target gids were already calculated
        for another target.
        """
        # Abort if there are no cx files yet or in case now we request full circuit
        # since its impossible to have a superset of it
        if (not target_spec.name or not self._cx_targets):
            logging.info(" => Target Cx reusing is not available.")
            return False

        logging.info("Attempt reusing cx files from other targets...")
        target_gids = self._get_target_gids(target_spec)
        cx_other = {}

        for previous_target in self._cx_targets:
            log_verbose("Trying previous cx file on target %s", previous_target)
            ref_filename = self._cx_filename(previous_target)
            if self._cx_contains_gids(ref_filename, target_gids, cx_other):
                break  # done!
            log_verbose("  - Target is not a superset. Ignoring.")
        else:
            logging.info(" => Did not find any suitable target")
            return False

        new_cx_filename = self._cx_filename(target_spec.simple_name)
        logging.info("Target %s is a subset of the target %s. Generating %s",
                     target_spec.name, previous_target, new_cx_filename)

        # Write the new cx file since Neuron needs it to do CPU assignment
        with open(new_cx_filename, "w") as newfile:
            self._write_msdat_dict(newfile, cx_other, target_gids)
        # register
        self._cx_targets.add(target_spec.simple_name)
        return True

    # -
    def _cx_valid(self, target_spec) -> bool:
        """Determine if valid complexity files exist for the provided circuit and
        target spec, otherwise return False -> we need to regen load balance info
        """
        # _cx_targets populated at instantiation. If not there, no chance. If in list, verify
        if not self._cx_targets:
            logging.info(" => No complexity files for current circuit yet")
            return False

        target_name = target_spec.simple_name
        cx_filename = self._cx_filename(target_name)

        if target_name not in self._cx_targets:
            logging.info(" => No Cx files available for requested target")
            return False

        if target_spec:  # target provided, otherwise everything
            target_gids = self._get_target_gids(target_spec)
            if not self._cx_contains_gids(cx_filename, target_gids):
                logging.warning(" => %s invalid: changed target definition!", cx_filename)
                return False
        return True

    @classmethod
    def _cx_contains_gids(cls, cxpath, target_gids, out_cx=None) -> bool:
        """Checks a cx file contains complexities for given gids
        """
        if not cxpath.is_file():
            log_verbose("  - cxpath doesnt exist: %s", cxpath)
            return False
        with open(cxpath, "r") as f:
            cx_saved = cls._read_msdat(f)
        if not set(cx_saved.keys()) >= set(target_gids):
            log_verbose("  - Not all GIDs in target %s %s", set(cx_saved.keys()), set(target_gids))
            return False
        if out_cx is not None:
            out_cx.update(cx_saved)
        return True

    @contextmanager
    def generate_load_balance(self, target_spec, cell_distributor):
        """Context manager that creates load balance for the circuit instantiated within

        Args:
            target_str: a string representation of the target.
            cell_distributor: the cell distributor object to which we can query
                the cells to be load balanced
        """
        mcomplex = Nd.MComplexLoadBalancer()  # init mcomplex before building circuit
        yield
        target_str = target_spec.simple_name
        self._compute_save_complexities(target_str, mcomplex, cell_distributor)
        self._cpu_assign(target_str)
        self._valid_loadbalance.add(target_str)

    @mpi_no_errors
    def _compute_save_complexities(self, target_str, mcomplex, cell_distributor):
        msfactor = 1e6 if self.lb_mode == LoadBalanceMode.WholeCell else 0.8
        out_filename = self._cx_filename(target_str)

        cx_cells = self._compute_complexities(mcomplex, cell_distributor)
        total_cx, max_cx = self._cell_complexity_total_max(cx_cells)
        lcx = self._get_optimal_piece_complexity(total_cx,
                                                 self.target_cpu_count,
                                                 msfactor)
        logging.info("LB Info: TC=%.3f MC=%.3f OptimalCx=%.3f FileName=%s",
                     total_cx, max_cx, lcx, out_filename)

        ms_list = []
        tmp = Nd.Vector()

        for cell in cell_distributor.cells:
            mcomplex.cell_complexity(cell.CellRef)
            mcomplex.multisplit(cell.gid, lcx, tmp)
            ms_list.append(tmp.c())

        # To output build independently the contents of the file then append
        ostring = StringIO()
        for ms in ms_list:
            self._write_msdat(ostring, ms)

        all_ranks_cx = MPI.py_gather(ostring.getvalue(), 0)
        if MPI.rank == 0:
            with open(out_filename, "w") as fp:
                fp.write("1\n%d\n" % cell_distributor.total_cells)
                for cx_info in all_ranks_cx:
                    fp.write(cx_info)
        # register
        self._cx_targets.add(target_str)

    @staticmethod
    def _cell_complexity_total_max(cx_cells):
        """
        Returns: Tuple of (TotalComplexity, max_complexity)
        """
        local_max = max(cx_cells) if len(cx_cells) > 0 else .0
        local_sum = sum(cx_cells) if len(cx_cells) > 0 else .0

        global_total = MPI.allreduce(local_sum, MPI.SUM)
        global_max = MPI.allreduce(local_max, MPI.MAX)
        return global_total, global_max

    @classmethod
    def _compute_complexities(cls, mcomplex, cell_distributor):
        cx_cell = compat.Vector("f")
        pc = cell_distributor.pc
        for gid in cell_distributor.local_nodes.final_gids():
            cx_cell.append(mcomplex.cell_complexity(pc.gid2cell(gid)))
        return cx_cell

    @staticmethod
    def _get_optimal_piece_complexity(total_cx, nhost, msfactor):
        """
        Args:
            total_cx: Total complexity
            nhost: Prospective no of hosts
        """
        lps = total_cx * msfactor / nhost
        return int(lps+1)

    @run_only_rank0
    def _cpu_assign(self, target_name):
        """Assigns cells to 'prospective_hosts' cpus using mymetis3.
        Results are written to file. basename.<NCPU>.dat
        """
        logging.info("Assigning Cells <-> %d CPUs [mymetis3]", self.target_cpu_count)
        base_filename = self._cx_filename(target_name, True)
        Nd.mymetis3(base_filename, self.target_cpu_count)

    @staticmethod
    def _write_msdat(fp, ms):
        """Writes load balancing info to an output stream
        """
        fp.write("%d" % ms.x[0])   # gid
        fp.write(" %g" % ms.x[1])  # total complexity of cell
        piece_count = int(ms.x[2])
        fp.write(" %d\n" % piece_count)
        i = 2
        tcx = 0  # Total accum complexity

        for _ in range(piece_count):
            i += 1
            subtree_count = int(ms.x[i])
            fp.write("  %d\n" % subtree_count)
            for _ in range(subtree_count):
                i += 1
                cx = ms.x[i]  # subtree complexity
                tcx += cx
                i += 1
                children_count = int(ms.x[i])
                fp.write("   %g %d\n" % (cx, children_count))
                if children_count > 0:
                    fp.write("    ")
                for _ in range(children_count):
                    i += 1
                    elem_id = ms.x[i]  # at next child
                    fp.write(" %d" % elem_id)
                if children_count > 0:
                    fp.write("\n")

    @staticmethod
    def _read_msdat(fp):
        """read load balancing info from an input stream
        """
        cx_saved = {}  # dict with key = gid, value = line content
        piece_count = 0
        gid = None
        next(fp)  # first line has 1. skip
        next(fp)  # ncells. skip

        for line in fp:
            if piece_count == 0:
                gid, _cx, piece_count = [int(float(x)) for x in line.split()]
                cx_saved[gid] = [line]
            else:                               # Handle parts
                cx_saved[gid].append(line)
                for _ in range(2 * int(line)):  # each subtree has two lines
                    cx_saved[gid].append(next(fp))
                piece_count -= 1

        return cx_saved

    @staticmethod
    def _write_msdat_dict(fp, cx_dict, gids=None):
        """Write out selected gid cx lines from a cx_dict
        """
        if gids is None:
            gids = cx_dict.keys()
        fp.write("1\n%d\n" % len(gids))
        for gid in gids:
            for line in cx_dict[gid]:
                fp.write(line)  # raw lines, include \n

    # -
    def _get_target_gids(self, target_spec) -> numpy.ndarray:
        return self._target_manager.get_target(target_spec).get_gids()

    def load_balance_info(self, target_spec):
        """ Loads a load-balance info for a given target.
        NOTE: Please ensure the load balance exists or is derived before calling this function
        """
        bal_filename = self._cx_filename(target_spec.simple_name, True)
        return Nd.BalanceInfo(bal_filename, MPI.rank, MPI.size)

    @classmethod
    def _loadbal_dir(cls, nodefile, population) -> Path:
        """Returns the dir where load balance files are stored for a given nodes file"""
        nodefile_hash = hashlib.md5(nodefile.encode()).digest().hex()[:10]
        return Path(cls._base_output_dir) / (cls._circuit_lb_dir_tpl % (nodefile_hash, population))

    def _cx_filename(self, target_str, basename_str=False) -> Path:
        """Gets the filename of a cell complexity file for a given target"""
        fname = self._lb_dir / (self._cx_filename_tpl % target_str)
        return str(fname)[:-4] if basename_str else fname

    def _cpu_assign_filename(self, target_str) -> Path:
        """Gets the CPU assignment filename for a given target, according to target CPU count"""
        return self._lb_dir / (self._cpu_assign_filename_tpl % (target_str, self.target_cpu_count))

    @staticmethod
    def select_lb_mode(sim_config, run_conf, target):
        """ A method which selects the load balance mode according to run config
        """
        # Check / set load balance mode
        lb_mode = sim_config.loadbal_mode
        if lb_mode == LoadBalanceMode.MultiSplit:
            if not sim_config.use_coreneuron:
                logging.info("Load Balancing ENABLED. Mode: MultiSplit")
            else:
                logging.warning("Load Balancing mode CHANGED to WholeCell for CoreNeuron")
                lb_mode = LoadBalanceMode.WholeCell

        elif lb_mode == LoadBalanceMode.WholeCell:
            logging.info("Load Balancing ENABLED. Mode: WholeCell")

        elif lb_mode is None:
            if target.is_void():
                lb_mode, reason = LoadBalanceMode.RoundRobin, "No target set, unknown cell count"
            else:
                lb_mode, reason = LoadBalanceMode.auto_select(sim_config.use_neuron,
                                                              target.gid_count(),
                                                              run_conf["Duration"])
            logging.warning("Load Balance AUTO-SELECTED: %s. Reason: %s", lb_mode.name, reason)

        return lb_mode
